import abc
import typing
from dataclasses import field

from pydantic.dataclasses import dataclass

from celus_nibbler.conditions import Condition
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.non_counter.celus_format import (
    BaseCelusFormatArea,
    BaseCelusFormatParser,
)
from celus_nibbler.sources import ExtractParams, SpecialExtraction
from celus_nibbler.utils import JsonEncorder, PydanticConfig
from celus_nibbler.validators import validators

from .base import BaseAreaDefinition, BaseNonCounterParserDefinition, ValidatorChoices


@dataclass(config=PydanticConfig)
class CelusFormatAreaDefinition(JsonEncorder, BaseAreaDefinition):
    title_column_names: typing.List[str] = field(default_factory=lambda: [])
    item_column_names: typing.List[str] = field(default_factory=lambda: [])
    organization_column_names: typing.List[str] = field(default_factory=lambda: [])
    metric_column_names: typing.List[str] = field(default_factory=lambda: [])
    default_metric: typing.Optional[str] = None
    title_ids_mapping: typing.Dict[str, str] = field(default_factory=lambda: {})
    item_ids_mapping: typing.Dict[str, str] = field(default_factory=lambda: {})
    item_publication_date_column_names: typing.List[str] = field(default_factory=lambda: [])
    item_authors_column_names: typing.List[str] = field(default_factory=lambda: [])
    dimension_mapping: typing.Dict[str, str] = field(default_factory=lambda: {})
    value_extract_params: ExtractParams = field(default_factory=lambda: ExtractParams())
    aggregate_same_records: bool = False

    kind: typing.Literal["non_counter.celus_format"] = "non_counter.celus_format"

    def make_area(self):
        class Area(BaseCelusFormatArea):
            title_column_names = self.title_column_names
            item_column_names = self.item_column_names
            organization_column_names = self.organization_column_names
            metric_column_names = self.metric_column_names
            default_metric = self.default_metric
            title_ids_mapping = self.title_ids_mapping
            item_ids_mapping = self.item_ids_mapping
            item_publication_date_column_names = self.item_publication_date_column_names
            item_authors_column_names = self.item_authors_column_names
            dimension_mapping = self.dimension_mapping
            value_extract_params = self.value_extract_params
            aggregator = self.make_aggregator()

        return Area


@dataclass(config=PydanticConfig)
class CelusFormatParserDefinition(BaseNonCounterParserDefinition, metaclass=abc.ABCMeta):
    parser_name: str
    data_format: DataFormatDefinition
    areas: typing.List[CelusFormatAreaDefinition]

    platforms: typing.List[str] = field(default_factory=lambda: [])
    metrics_to_skip: typing.List[str] = field(default_factory=lambda: [])
    available_metrics: typing.Optional[typing.List[str]] = None
    on_metric_check_failed: TableException.Action = TableException.Action.SKIP
    titles_to_skip: typing.List[str] = field(default_factory=lambda: [])
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = field(default_factory=lambda: {})
    dimensions_validators: typing.Dict[str, ValidatorChoices] = field(default_factory=lambda: {})
    metric_aliases: typing.List[typing.Tuple[str, str]] = field(default_factory=lambda: [])
    metric_value_extraction_overrides: typing.Dict[str, SpecialExtraction] = field(
        default_factory=lambda: {}
    )
    dimension_aliases: typing.List[typing.Tuple[str, str]] = field(default_factory=lambda: [])
    heuristics: typing.Optional[Condition] = None
    possible_row_offsets: typing.List[int] = field(default_factory=lambda: [0])

    kind: typing.Literal["non_counter.celus_format"] = "non_counter.celus_format"
    version: typing.Literal[1] = 1

    def make_parser(self):
        validators_map = {e.name: e for e in validators}
        _dimensions_validators = {
            dimension_name: validators_map[validator_name]
            for dimension_name, validator_name in self.dimensions_validators.items()
        }

        class Parser(BaseCelusFormatParser):
            _definition = self
            data_format = _definition.data_format
            name = (
                f"dynamic.{_definition.group}.{_definition.data_format.name}"
                f".{_definition.parser_name}"
            )

            platforms = _definition.platforms
            metrics_to_skip = self.metrics_to_skip
            titles_to_skip = self.titles_to_skip
            dimensions_to_skip = self.dimensions_to_skip
            dimensions_validators = _dimensions_validators
            available_metrics = self.available_metrics
            on_metric_check_failed = self.on_metric_check_failed

            metric_aliases = dict(self.metric_aliases)
            metric_value_extraction_overrides = dict(self.metric_value_extraction_overrides)
            dimension_aliases = dict(self.dimension_aliases)

            heuristics = self.heuristics
            possible_row_offsets = self.possible_row_offsets

            areas = [e.make_area() for e in self.areas]

        return Parser

    kind: typing.Literal["non_counter.celus_format"] = "non_counter.celus_format"



import collections
import itertools
import re
import sys
import typing
from importlib.metadata import entry_points

from .base import BaseParser

PYTHON_VERSION = sys.version_info

if PYTHON_VERSION[0] == 3 and PYTHON_VERSION[1] < 10:

    def get_entry_points():
        return entry_points().get("nibbler_parsers", [])

else:

    def get_entry_points():
        return entry_points(group="nibbler_parsers")


def get_parsers(
    parsers: typing.Optional[typing.List[str]] = None,
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> typing.List[typing.Tuple[str, typing.Type[BaseParser]]]:
    """Lists all (name, parser) tuples

    :param parsers: use only parsers which names are on this list
    :returns: List of tuples with (name, parser class)
    """
    return [
        (entry_point.name, entry_point.load())
        for entry_point in get_entry_points()
        if not parsers or any(re.match(e, entry_point.name) for e in parsers)
    ] + [
        (parser.name, parser)
        for parser in dynamic_parsers
        if not parsers or any(re.match(e, parser.name) for e in parsers)
    ]


def available_parsers(
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> typing.List[str]:
    return [name for name, _ in get_parsers(dynamic_parsers=dynamic_parsers)]


def filter_parsers(
    parsers: typing.Optional[typing.List[str]] = None,
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> typing.List[typing.Type[BaseParser]]:
    """Lists all available parsers

    :param parsers: use only parsers which names are on this list
    :returns: List of parser classes
    """
    return [entry_point for _, entry_point in get_parsers(parsers, dynamic_parsers)]


def get_supported_platforms(
    parsers: typing.Optional[typing.List[str]] = None,
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> typing.List[str]:
    """Lists all supported platforms"""
    return sorted(
        list(
            set(
                itertools.chain.from_iterable(
                    parser.platforms for parser in filter_parsers(parsers, dynamic_parsers)
                )
            )
        )
    )


def get_supported_platforms_count(
    parsers: typing.Optional[typing.List[str]] = None,
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> typing.List[typing.Tuple[str, int]]:
    counter: typing.Dict[str, int] = collections.Counter()
    for parser in filter_parsers(parsers, dynamic_parsers):
        for platform in parser.platforms:
            counter[platform] += 1

    res = []
    for key in sorted(counter.keys()):
        res.append((key, counter[key]))

    return res


__all__ = [
    "filter_parsers",
    "get_supported_platforms",
    "get_supported_platforms_count",
]



import abc
import typing
from dataclasses import field

from pydantic import Field
from pydantic.dataclasses import dataclass
from typing_extensions import Annotated

from celus_nibbler.conditions import Condition
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.base import BaseArea, BaseParser
from celus_nibbler.parsers.counter import CounterHeaderArea, c51
from celus_nibbler.parsers.counter.c4 import (
    BR1,
    BR2,
    BR3,
    DB1,
    DB2,
    JR1,
    JR1GOA,
    JR2,
    MR1,
    PR1,
    BaseCounter4Parser,
    JR1a,
)
from celus_nibbler.parsers.counter.c5 import DR, IR, IR_M1, PR, TR, BaseCounter5Parser
from celus_nibbler.sources import ExtractParams, SpecialExtraction
from celus_nibbler.utils import JsonEncorder, PydanticConfig
from celus_nibbler.validators import validators

from .base import BaseAreaDefinition, BaseParserDefinition, ValidatorChoices


@dataclass(config=PydanticConfig)
class BaseCounterAreaDefinition(BaseAreaDefinition):
    kind: str = NotImplementedError
    organization_column: typing.Optional[typing.List[str]] = None
    metric_column: typing.Optional[typing.List[str]] = None
    title_column: typing.Optional[typing.List[str]] = None
    item_column: typing.Optional[typing.List[str]] = None
    organization_extract_params: typing.Optional[ExtractParams] = None
    metric_extract_params: typing.Optional[ExtractParams] = None
    date_extract_params: typing.Optional[ExtractParams] = None
    data_extract_params: typing.Optional[ExtractParams] = None
    title_extract_params: typing.Optional[ExtractParams] = None
    title_ids_extract_params: typing.Optional[typing.Dict[str, ExtractParams]] = None
    item_extract_params: typing.Optional[ExtractParams] = None
    item_ids_extract_params: typing.Optional[typing.Dict[str, ExtractParams]] = None
    dimensions_extract_params: typing.Optional[typing.Dict[str, ExtractParams]] = None
    aggregate_same_records: bool = False

    def make_area(self):
        raise NotImplementedError()


def gen_area(
    base: typing.Type[BaseArea], definition: BaseCounterAreaDefinition
) -> typing.Type[CounterHeaderArea]:
    class Area(base):
        # Column names
        if definition.organization_column is not None:
            ORGANIZATION_COLUMN_NAMES = definition.organization_column
        if definition.metric_column is not None:
            METRIC_COLUMN_NAMES = definition.metric_column
        if definition.title_column is not None:
            TITLE_COLUMN_NAMES = definition.title_column
        if definition.item_column is not None:
            ITEM_COLUMN_NAMES = definition.item_column

        # Extra params
        if definition.organization_extract_params is not None:
            ORGANIZATION_EXTRACT_PARAMS = definition.organization_extract_params
        if definition.metric_extract_params is not None:
            METRIC_EXTRACT_PARAMS = definition.metric_extract_params
        if definition.date_extract_params is not None:
            DATE_EXTRACT_PARAMS = definition.date_extract_params
        if definition.data_extract_params is not None:
            DATA_EXTRACT_PARAMS = definition.data_extract_params
        if definition.title_extract_params is not None:
            TITLE_EXTRACT_PARAMS = definition.title_extract_params
        if definition.title_ids_extract_params is not None:
            TITLE_IDS_EXTRACT_PARAMS = definition.title_ids_extract_params
        if definition.item_extract_params is not None:
            ITEM_EXTRACT_PARAMS = definition.item_extract_params
        if definition.item_ids_extract_params is not None:
            ITEM_IDS_EXTRACT_PARAMS = definition.item_ids_extract_params
        if definition.dimensions_extract_params is not None:
            DIMENSIONS_EXTRACT_PARAMS = definition.dimensions_extract_params

        aggregator = definition.make_aggregator()

    return Area


@dataclass(config=PydanticConfig)
class BR1AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.BR1"] = "counter4.BR1"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(BR1.Area, self)


@dataclass(config=PydanticConfig)
class BR2AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.BR2"] = "counter4.BR2"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(BR2.Area, self)


@dataclass(config=PydanticConfig)
class BR3AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.BR3"] = "counter4.BR3"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(BR3.Area, self)


@dataclass(config=PydanticConfig)
class DB1AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.DB1"] = "counter4.DB1"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(DB1.Area, self)


@dataclass(config=PydanticConfig)
class DB2AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.DB2"] = "counter4.DB2"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(DB2.Area, self)


@dataclass(config=PydanticConfig)
class JR1AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.JR1"] = "counter4.JR1"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(JR1.Area, self)


@dataclass(config=PydanticConfig)
class JR1aAreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.JR1a"] = "counter4.JR1a"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(JR1a.Area, self)


@dataclass(config=PydanticConfig)
class JR1GOAAreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.JR1GOA"] = "counter4.JR1GOA"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(JR1GOA.Area, self)


@dataclass(config=PydanticConfig)
class JR2AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.JR2"] = "counter4.JR2"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(JR2.Area, self)


@dataclass(config=PydanticConfig)
class MR1AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.MR1"] = "counter4.MR1"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(MR1.Area, self)


@dataclass(config=PydanticConfig)
class PR1AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter4.PR1"] = "counter4.PR1"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(PR1.Area, self)


@dataclass(config=PydanticConfig)
class TRAreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter5.TR"] = "counter5.TR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(TR.Area, self)


@dataclass(config=PydanticConfig)
class IR_M1AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter5.IR_M1"] = "counter5.IR_M1"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(IR_M1.Area, self)


@dataclass(config=PydanticConfig)
class IRAreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter5.IR"] = "counter5.IR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(IR.Area, self)


@dataclass(config=PydanticConfig)
class DRAreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter5.DR"] = "counter5.DR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(DR.Area, self)


@dataclass(config=PydanticConfig)
class PRAreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter5.PR"] = "counter5.PR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(PR.Area, self)


@dataclass(config=PydanticConfig)
class TR51AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter51.TR"] = "counter51.TR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(c51.TR.Area, self)


@dataclass(config=PydanticConfig)
class IR51AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter51.IR"] = "counter51.IR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(c51.IR.Area, self)


@dataclass(config=PydanticConfig)
class DR51AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter51.DR"] = "counter51.DR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(c51.DR.Area, self)


@dataclass(config=PydanticConfig)
class PR51AreaDefinition(BaseCounterAreaDefinition):
    kind: typing.Literal["counter51.PR"] = "counter51.PR"

    def make_area(self) -> typing.Type[CounterHeaderArea]:
        return gen_area(c51.PR.Area, self)


CounterAreaDefinition = Annotated[
    typing.Union[
        BR1AreaDefinition,
        BR2AreaDefinition,
        BR3AreaDefinition,
        DB1AreaDefinition,
        DB2AreaDefinition,
        JR1AreaDefinition,
        JR1aAreaDefinition,
        JR1GOAAreaDefinition,
        JR2AreaDefinition,
        MR1AreaDefinition,
        PR1AreaDefinition,
        DRAreaDefinition,
        PRAreaDefinition,
        TRAreaDefinition,
        IR_M1AreaDefinition,
        IRAreaDefinition,
        DR51AreaDefinition,
        PR51AreaDefinition,
        TR51AreaDefinition,
        IR51AreaDefinition,
    ],
    Field(discriminator="kind"),
]


@dataclass(config=PydanticConfig)
class BaseCounterParserDefinition(BaseParserDefinition, metaclass=abc.ABCMeta):
    parser_name: str
    kind: str
    group: str
    heuristics: typing.Optional[Condition] = None
    data_format: typing.Optional[DataFormatDefinition] = None
    platforms: typing.Optional[typing.List[str]] = None
    areas: typing.List[CounterAreaDefinition] = field(
        default_factory=lambda: [], metadata={"min_items": 1, "max_items": 1}
    )
    metrics_to_skip: typing.List[str] = field(default_factory=lambda: [])
    available_metrics: typing.Optional[typing.List[str]] = None
    on_metric_check_failed: TableException.Action = TableException.Action.SKIP
    titles_to_skip: typing.List[str] = field(default_factory=lambda: [])
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = field(default_factory=lambda: {})
    dimensions_validators: typing.Dict[str, ValidatorChoices] = field(default_factory=lambda: {})
    metric_aliases: typing.List[typing.Tuple[str, str]] = field(default_factory=lambda: [])
    metric_value_extraction_overrides: typing.Dict[str, SpecialExtraction] = field(
        default_factory=lambda: {}
    )
    dimension_aliases: typing.List[typing.Tuple[str, str]] = field(default_factory=lambda: [])
    possible_row_offsets: typing.List[int] = field(default_factory=lambda: [0])


@dataclass(config=PydanticConfig)
class BaseCounter4ParserDefinition(BaseCounterParserDefinition):
    group: str = "counter4"


@dataclass(config=PydanticConfig)
class BaseCounter5ParserDefinition(BaseCounterParserDefinition):
    group: str = "counter5"


@dataclass(config=PydanticConfig)
class BaseCounter51ParserDefinition(BaseCounterParserDefinition):
    group: str = "counter51"


def gen_parser(
    base: typing.Union[typing.Type[BaseCounter4Parser], typing.Type[BaseCounter5Parser]],
    definition: BaseCounterParserDefinition,
) -> typing.Type[BaseParser]:
    validators_map = {e.name: e for e in validators}
    _dimensions_validators = {
        dimension_name: validators_map[validator_name]
        for dimension_name, validator_name in definition.dimensions_validators.items()
    }

    class Parser(base):
        data_format: DataFormatDefinition = (
            definition.data_format if definition.data_format else base.data_format
        )
        name = f"dynamic.{definition.group}.{data_format.name}.{definition.parser_name}"
        platforms = definition.platforms or base.platforms
        heuristics = definition.heuristics or base.heuristics
        metrics_to_skip = definition.metrics_to_skip or base.metrics_to_skip
        available_metrics = definition.available_metrics or base.available_metrics
        on_metric_check_failed = definition.on_metric_check_failed or base.on_metric_check_failed
        titles_to_skip = definition.titles_to_skip or base.titles_to_skip
        dimensions_to_skip = definition.dimensions_to_skip or base.dimensions_to_skip
        dimensions_validators = _dimensions_validators
        metric_aliases = dict(definition.metric_aliases) or dict(base.metric_aliases)
        metric_value_extraction_overrides = dict(definition.metric_value_extraction_overrides)
        dimension_aliases = dict(definition.dimension_aliases) or dict(base.dimension_aliases)
        possible_row_offsets = definition.possible_row_offsets or base.possible_row_offsets

        areas = [definition.areas[0].make_area()]

    return Parser


@dataclass(config=PydanticConfig)
class BR1Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.BR1"] = "counter4.BR1"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(BR1, self)


@dataclass(config=PydanticConfig)
class BR2Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.BR2"] = "counter4.BR2"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(BR2, self)


@dataclass(config=PydanticConfig)
class BR3Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.BR3"] = "counter4.BR3"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(BR3, self)


@dataclass(config=PydanticConfig)
class DB1Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.DB1"] = "counter4.DB1"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(DB1, self)


@dataclass(config=PydanticConfig)
class DB2Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.DB2"] = "counter4.DB2"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(DB2, self)


@dataclass(config=PydanticConfig)
class JR1Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.JR1"] = "counter4.JR1"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(JR1, self)


@dataclass(config=PydanticConfig)
class JR1aDefinition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.JR1a"] = "counter4.JR1a"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(JR1a, self)


@dataclass(config=PydanticConfig)
class JR1GOADefinition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.JR1GOA"] = "counter4.JR1GOA"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(JR1GOA, self)


@dataclass(config=PydanticConfig)
class JR2Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.JR2"] = "counter4.JR2"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(JR2, self)


@dataclass(config=PydanticConfig)
class MR1Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.MR1"] = "counter4.MR1"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(MR1, self)


@dataclass(config=PydanticConfig)
class PR1Definition(JsonEncorder, BaseCounter4ParserDefinition):
    kind: typing.Literal["counter4.PR1"] = "counter4.PR1"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(PR1, self)


@dataclass(config=PydanticConfig)
class TRDefinition(JsonEncorder, BaseCounter5ParserDefinition):
    kind: typing.Literal["counter5.TR"] = "counter5.TR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(TR, self)


@dataclass(config=PydanticConfig)
class DRDefinition(JsonEncorder, BaseCounter5ParserDefinition):
    kind: typing.Literal["counter5.DR"] = "counter5.DR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(DR, self)


@dataclass(config=PydanticConfig)
class PRDefinition(JsonEncorder, BaseCounter5ParserDefinition):
    kind: typing.Literal["counter5.PR"] = "counter5.PR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(PR, self)


@dataclass(config=PydanticConfig)
class IR_M1Definition(JsonEncorder, BaseCounter5ParserDefinition):
    kind: typing.Literal["counter5.IR_M1"] = "counter5.IR_M1"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(IR_M1, self)


@dataclass(config=PydanticConfig)
class IRDefinition(JsonEncorder, BaseCounter5ParserDefinition):
    kind: typing.Literal["counter5.IR"] = "counter5.IR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(IR, self)


@dataclass(config=PydanticConfig)
class TR51Definition(JsonEncorder, BaseCounter51ParserDefinition):
    kind: typing.Literal["counter51.TR"] = "counter51.TR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(c51.TR, self)


@dataclass(config=PydanticConfig)
class DR51Definition(JsonEncorder, BaseCounter51ParserDefinition):
    kind: typing.Literal["counter51.DR"] = "counter51.DR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(c51.DR, self)


@dataclass(config=PydanticConfig)
class PR51Definition(JsonEncorder, BaseCounter51ParserDefinition):
    kind: typing.Literal["counter51.PR"] = "counter51.PR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(c51.PR, self)


@dataclass(config=PydanticConfig)
class IR51Definition(JsonEncorder, BaseCounter51ParserDefinition):
    kind: typing.Literal["counter51.IR"] = "counter51.IR"
    data_format: typing.Optional[DataFormatDefinition] = None
    version: typing.Literal[1] = 1

    def make_parser(self):
        return gen_parser(c51.IR, self)


CounterDefinition = Annotated[
    typing.Union[
        BR1Definition,
        BR2Definition,
        BR3Definition,
        DB1Definition,
        DB2Definition,
        JR1Definition,
        JR1aDefinition,
        JR1GOADefinition,
        JR2Definition,
        MR1Definition,
        PR1Definition,
        DRDefinition,
        PRDefinition,
        TRDefinition,
        IR_M1Definition,
        IRDefinition,
        DR51Definition,
        PR51Definition,
        TR51Definition,
        IR51Definition,
    ],
    Field(discriminator="kind"),
]



import copy
import datetime
import logging
import typing

from celus_nibbler.coordinates import Coord, CoordRange, Direction, RelativeTo, Value
from celus_nibbler.data_headers import DataHeaders
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.base import BaseHeaderArea, BaseTabularParser
from celus_nibbler.sources import (
    AuthorsSource,
    DateSource,
    DimensionSource,
    ExtractParams,
    ItemIdSource,
    ItemSource,
    MetricSource,
    OrganizationSource,
    PublicationDateSource,
    TitleIdSource,
    TitleSource,
)

from .base import BaseNonCounterParser

logger = logging.getLogger(__name__)


class BaseCelusFormatArea(BaseHeaderArea):
    title_column_names: typing.List[str] = []
    item_column_names: typing.List[str] = []
    organization_column_names: typing.List[str] = []
    metric_column_names: typing.List[str] = []
    default_metric: typing.Optional[str] = None
    title_ids_mapping: typing.Dict[str, str] = {}
    item_ids_mapping: typing.Dict[str, str] = {}
    item_publication_date_column_names: typing.List[str] = []
    item_authors_column_names: typing.List[str] = []
    dimension_mapping: typing.Dict[str, str] = {}
    value_extract_params: ExtractParams = ExtractParams()

    @property
    def data_headers(self) -> DataHeaders:
        # Should raise validation error when column is unidentified

        # Convert title_ids_sources and dimensions_sources to instance variables
        # otherwise class variable would be modified here causing
        # it to influence other parsers
        self.title_ids_sources = {}
        self.dimensions_sources = {}
        self.item_ids_sources = {}
        self.item_publication_date_source = None
        self.item_authors_source = None

        date_source = None
        for cell in CoordRange(Coord(0, 0, RelativeTo.START), Direction.RIGHT):
            data_source = CoordRange(cell, Direction.DOWN).skip(1)
            try:
                # First try to extract date
                try:
                    test_date_source = DateSource(CoordRange(cell, Direction.RIGHT))
                    if test_date_source.extract(
                        self.sheet, idx=0, parser_row_offset=0, area_row_offset=0
                    ):
                        # Set date source based on the first found date
                        if not date_source:
                            date_source = test_date_source
                        continue
                except TableException as e:
                    if e.reason != "cant-parse-date":
                        raise

                content = cell.content(self.sheet)

                # Try title
                if content in self.title_column_names:
                    self.title_source = TitleSource(source=data_source)
                    continue

                # Try item
                if content in self.item_column_names:
                    self.item_source = ItemSource(source=data_source)
                    continue

                # Try metric
                if content in self.metric_column_names:
                    self.metric_source = MetricSource(source=data_source)
                    continue

                # Try organization
                if content in self.organization_column_names:
                    self.organization_source = OrganizationSource(source=data_source)
                    continue

                # Try item authors
                if content in self.item_authors_column_names:
                    self.item_authors_source = AuthorsSource(source=data_source)
                    continue

                # Try item publication date
                if content in self.item_publication_date_column_names:
                    self.item_publication_date_source = PublicationDateSource(source=data_source)
                    continue

                # Try dimensions
                if name := self.dimension_mapping.get(content):
                    self.dimensions_sources[name] = DimensionSource(name=name, source=data_source)
                    continue

                # Try title ids
                if name := self.title_ids_mapping.get(content):
                    self.title_ids_sources[name] = TitleIdSource(name=name, source=data_source)
                    continue

                # Try title ids
                if name := self.item_ids_mapping.get(content):
                    self.item_ids_sources[name] = ItemIdSource(name=name, source=data_source)
                    continue

                # Unknown column
                raise TableException(
                    value=content,
                    row=cell.row,
                    col=cell.col,
                    sheet=self.sheet.sheet_idx,
                    reason="unknown-column",
                )
            except TableException as e:
                # Stop processing when an exception occurs
                # (index out of bounds or unable to parse next field)
                if e.action == TableException.Action.STOP:
                    logger.debug("Header parsing terminated: %s", e)
                    break
                else:
                    raise

        if self.default_metric:
            if self.metric_source:
                # Need to clone here to copy&modify class variable
                # to instance variable
                cloned = copy.deepcopy(self.metric_source)
                cloned.extract_params.default = self.default_metric
                self.metric_source = cloned
            else:
                self.metric_source = MetricSource(Value(self.default_metric))

        # Raise exception when first row doesn't contain any date
        if not date_source:
            raise TableException(
                value=None,
                row=0,
                col=None,
                sheet=self.sheet.sheet_idx,
                reason="no-date-in-header",
            )

        # data cells are right under date_source
        data_cells: CoordRange = copy.deepcopy(date_source.source)
        data_cells.coord.row += 1

        return DataHeaders(
            roles=[date_source],
            data_cells=data_cells,
            data_direction=Direction.DOWN,
            data_extract_params=self.value_extract_params,
        )

    def dimensions(self) -> typing.List[str]:
        # populate dimensions_sources
        self.find_data_cells(self.sheet, lambda x: x, lambda x, y: None)
        return list(self.dimensions_sources.keys())

    def get_months(self) -> typing.List[datetime.date]:
        return self._get_months_from_header()


class BaseCelusFormatParser(BaseNonCounterParser, BaseTabularParser):
    """Fist row of generic celus format is the header and the rest are data"""

    platforms: typing.List[str] = []
    metrics_to_skip: typing.List[str] = []
    titles_to_skip: typing.List[str] = []
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = {}

    heuristics = None  # Celus format has no heuristcs



from .eat_and_poop import Poop, PoopOrganizationStats, PoopStats, eat
from .errors import (
    MultipleParsersFound,
    NibblerError,
    NoParserForPlatformFound,
    NoParserFound,
    NoParserMatchesHeuristics,
    RecordError,
    TableException,
    WrongFileFormatError,
)
from .parsers import get_supported_platforms

__all__ = [
    "eat",
    "get_supported_platforms",
    "Poop",
    "PoopStats",
    "PoopOrganizationStats",
    "MultipleParsersFound",
    "NibblerError",
    "NoParserFound",
    "NoParserForPlatformFound",
    "NoParserMatchesHeuristics",
    "RecordError",
    "TableException",
    "WrongFileFormatError",
]



import logging
from datetime import date
from typing import Generator, List

from celus_nigiri import CounterRecord
from celus_nigiri.counter5 import (
    Counter5DRReport,
    Counter5IRM1Report,
    Counter5IRReport,
    Counter5PRReport,
    Counter5ReportBase,
    Counter5TRReport,
)
from celus_nigiri.utils import get_date_range
from pydantic import ValidationError

from celus_nibbler import validators
from celus_nibbler.conditions import SheetExtraCondition
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.parsers.base import BaseArea, BaseJsonArea, BaseJsonParser
from celus_nibbler.reader import JsonCounter5SheetReader

from . import c5 as c5tabular

logger = logging.getLogger(__name__)


class NigiriBaseArea(BaseJsonArea):
    nigiri_report_class = Counter5ReportBase

    def get_months(self) -> List[date]:
        months_str = set()
        if report_filters := self.sheet.extra.get("Report_Filters"):
            begin = None
            end = None
            for rf in report_filters:
                if rf.get("Name") == "Begin_Date":
                    begin = rf.get("Value")
                if rf.get("Name") == "End_Date":
                    end = rf.get("Value")

            if begin and end:
                try:
                    begin_date = validators.Date(value=begin).value
                    end_date = validators.Date(value=end).value
                    return get_date_range(begin_date, end_date)
                except ValidationError:
                    logger.warn("Wrong date in Report_Filters")

        # No Report Filters detected => try to extract data from performance
        for item_dict in self.sheet:
            for permformance_item in item_dict.get("Performance", []):
                if period := permformance_item.get("Period"):
                    if begin_date := period.get("Begin_Date"):
                        months_str.add(begin_date)

        months = set()
        for month_str in months_str:
            try:
                months.add(validators.Date(value=month_str).value)
            except ValidationError:
                logger.warn("Wrong date in Report_Item: '%s'", month_str)

        return sorted(list(months))

    @property
    def dimensions(self) -> List[str]:
        return self.nigiri_report_class.dimensions


class NigiriDRArea(NigiriBaseArea):
    nigiri_report_class = Counter5DRReport


class NigiriPRArea(NigiriBaseArea):
    nigiri_report_class = Counter5PRReport


class NigiriTRArea(NigiriBaseArea):
    nigiri_report_class = Counter5TRReport


class NigiriIR_M1Area(NigiriBaseArea):
    nigiri_report_class = Counter5IRM1Report


class NigiriIRArea(NigiriBaseArea):
    nigiri_report_class = Counter5IRReport


class BaseCounter5JsonParser(c5tabular.Counter5ParserAnalyzeMixin, BaseJsonParser):
    @property
    def name(self):
        return f"static.counter5.{self.data_format.name}.Json"

    def _parse_area(self, area: BaseArea) -> Generator[CounterRecord, None, None]:
        if isinstance(area, NigiriBaseArea):
            report = area.nigiri_report_class()
            if isinstance(self.sheet, JsonCounter5SheetReader):
                return report.read_report(self.sheet.extra, self.sheet)
            raise TypeError(f"Only JsonCounter5SheetReader is allowed to be used in {type(self)}")
        raise TypeError(f"Only NigiriArea is allowed to be used in {type(self)}")

    def get_extras(self):
        return self.sheet.extra or {}


class DR(BaseCounter5JsonParser):
    heuristics = SheetExtraCondition(field_name="Report_ID", value="DR") & SheetExtraCondition(
        field_name="Release", value="5"
    )
    platforms = c5tabular.DR.platforms
    data_format = DataFormatDefinition(name="DR")

    areas = [NigiriDRArea]


class PR(BaseCounter5JsonParser):
    heuristics = SheetExtraCondition(field_name="Report_ID", value="PR") & SheetExtraCondition(
        field_name="Release", value="5"
    )
    platforms = c5tabular.PR.platforms
    data_format = DataFormatDefinition(name="PR")

    areas = [NigiriPRArea]


class TR(BaseCounter5JsonParser):
    heuristics = SheetExtraCondition(field_name="Report_ID", value="TR") & SheetExtraCondition(
        field_name="Release", value="5"
    )
    platforms = c5tabular.TR.platforms
    data_format = DataFormatDefinition(name="TR")

    areas = [NigiriTRArea]


class IR_M1(BaseCounter5JsonParser):
    heuristics = SheetExtraCondition(field_name="Report_ID", value="IR_M1") & SheetExtraCondition(
        field_name="Release", value="5"
    )
    platforms = c5tabular.IR_M1.platforms
    data_format = DataFormatDefinition(name="IR_M1")

    areas = [NigiriIR_M1Area]


class IR(BaseCounter5JsonParser):
    heuristics = SheetExtraCondition(field_name="Report_ID", value="IR") & SheetExtraCondition(
        field_name="Release", value="5"
    )
    platforms = c5tabular.IR.platforms
    data_format = DataFormatDefinition(name="IR")

    areas = [NigiriIRArea]



import argparse
import csv
import json
import logging
import logging.config
import pathlib
import sys
import typing
from contextlib import nullcontext
from copy import deepcopy
from datetime import date
from importlib.metadata import distribution

from celus_nigiri import CounterRecord
from unidecode import unidecode

from celus_nibbler import Poop, eat
from celus_nibbler.aggregator import CounterOrdering
from celus_nibbler.definitions import Definition
from celus_nibbler.parsers import available_parsers
from celus_nibbler.parsers.dynamic import gen_parser
from celus_nibbler.utils import profile


def read_definition(path: str) -> Definition:
    with pathlib.Path(path).open() as f:
        definition_data = json.load(f)
    return Definition.parse(definition_data)


def gen_description(parsers: typing.List[str]) -> str:
    parsers_list = "\n".join([f"  {e}" for e in parsers])

    return f"""Available Parsers:
{parsers_list}
"""


def write_batch(
    writer,
    records: typing.List[CounterRecord],
    dimensions: typing.List[str],
    title_ids: typing.List[str],
    item_ids: typing.List[str],
    months: typing.List[str],
):
    months_dict = {e: "" for e in months}
    for r in records:
        months_dict[r.start.strftime("%Y-%m")] = r.value
    writer.writerow(
        [r.organization, r.title]
        + [r.title_ids.get(ti, "") for ti in title_ids]
        + [r.item]
        + [r.item_ids.get(ii, "") for ii in item_ids]
        + [r.dimension_data.get(d, "") for d in dimensions]
        + [r.metric]
        + [months_dict[k] for k in sorted(months)]
    )


def gen_argument_parser(parsers: typing.List[str]) -> argparse.ArgumentParser:
    dist = distribution("celus-nibbler")
    version = dist.version if dist else "?"
    description = gen_description(parsers)
    parser = argparse.ArgumentParser(
        prog="nibbler-eat",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=description,
    )
    parser.add_argument("-d", "--debug", dest="debug", action="store_true", default=False)
    parser.add_argument("--version", action="version", version=version)
    parser.add_argument(
        "-s",
        "--skip-heuristics",
        dest="skip_heuristics",
        action="store_true",
        default=False,
        help="skip heuristic checks",
    )
    parser.add_argument(
        "-p",
        "--parser",
        action="append",
        default=[],
        help="limit parsers which can be used",
    )
    parser.add_argument(
        "-S",
        "--show-summary",
        action="store_true",
        default=False,
        help="show per sheet summary",
    )
    parser.add_argument(
        "-D",
        "--definition",
        action="append",
        default=[],
        help="Reads a definition of a dynamic parser from a file",
        type=read_definition,
    )
    parser.add_argument(
        "-P",
        "--platform",
        default="",
        help="Limit platform that can be used",
    )
    parser.add_argument(
        "-c",
        "--counter-like-output",
        action="store_true",
        default=False,
        help="Display processed data in counter like format",
    )
    parser.add_argument(
        "-N",
        "--no-output",
        action="store_true",
        default=False,
        help="Parses the entire file without producing any data output (useful for benchmarks)",
    )
    parser.add_argument("file", nargs="*")
    parser.add_argument("--profile", dest="profile", action="store_true", default=False)

    return parser


def parse(options, platform, dynamic_parsers):
    for file in options.file:
        if poops := eat(
            pathlib.Path(file),
            platform or "void",
            parsers=options.parser or None,
            check_platform=bool(platform),
            use_heuristics=not options.skip_heuristics,
            dynamic_parsers=dynamic_parsers,
        ):
            for idx, poop in enumerate(poops):
                if not isinstance(poop, Poop):
                    print(f"Failed to pick parser for sheet {idx}", file=sys.stderr)
                    continue

                stat_dict = poop.get_stats().dict()
                if options.show_summary:
                    print(f"Parsing sheet {idx}", file=sys.stderr)
                    print(f"Months: {stat_dict['months']}", file=sys.stderr)
                    print(f"Metrics: {stat_dict['metrics']}", file=sys.stderr)
                    print(f"Dimensions: {stat_dict['dimensions']}", file=sys.stderr)
                    print(f"Title ids: {stat_dict['title_ids']}", file=sys.stderr)
                    print(f"Item ids: {stat_dict['item_ids']}", file=sys.stderr)

                if options.no_output:
                    header = None
                elif options.counter_like_output:
                    header = (
                        ["Institution_Name", "Title"]
                        + list(poop.title_ids)
                        + list(poop.dimensions)
                        + ["Metric_Type"]
                        + list(poop.months)
                    )
                else:
                    header = [
                        "start",
                        "end",
                        "organization",
                        "title",
                        "title_ids",
                        "item",
                        "item_ids",
                        "item_publication_date",
                        "item_authors",
                        "dimensions",
                        "metric",
                        "value",
                    ]

                writer = csv.writer(sys.stdout, dialect="unix")
                if header:
                    writer.writerow(header)
                records = (e[1] for e in poop.records_basic())
                records = (
                    CounterOrdering().aggregate(records) if options.counter_like_output else records
                )
                last_rec = None
                batch = []
                for record in records:
                    if options.no_output:
                        # Just go through the records and suppress any output
                        pass
                    elif options.counter_like_output:
                        this_rec = deepcopy(record)
                        this_rec.value = -1
                        this_rec.start = date(2020, 1, 1)
                        this_rec.end = date(2020, 1, 31)

                        if not last_rec or last_rec == this_rec:
                            batch.append(record)
                        elif batch:
                            write_batch(
                                writer,
                                batch,
                                stat_dict["dimensions"].keys(),
                                stat_dict["title_ids"],
                                stat_dict["item_ids"],
                                stat_dict["months"].keys(),
                            )
                            batch = [record]
                        last_rec = this_rec
                    else:
                        writer.writerow(record.as_csv())

                if batch:
                    write_batch(
                        writer,
                        batch,
                        stat_dict["dimensions"].keys(),
                        stat_dict["title_ids"],
                        stat_dict["item_ids"],
                        stat_dict["months"].keys(),
                    )


def main():
    parser = gen_argument_parser(available_parsers())
    options = parser.parse_args()

    # init logging
    logging.basicConfig(level=logging.DEBUG if options.debug else logging.WARNING)
    logger = logging.getLogger(__name__)
    logger.debug("Logging is configured.")

    platform = unidecode(options.platform)
    dynamic_parsers = [gen_parser(e) for e in options.definition]

    if not options.file:
        # Just print help if not parsers specified
        # Enrich with dynamic parsers
        parser = gen_argument_parser(available_parsers(dynamic_parsers))
        parser.print_help()
        return

    context_man = nullcontext if not options.profile else profile

    with context_man():
        parse(options, platform, dynamic_parsers)


if __name__ == "__main__":
    main()



import abc
import typing
from enum import Enum

from pydantic.dataclasses import dataclass

from .errors import TableException
from .reader import SheetReader
from .utils import JsonEncorder, PydanticConfig


class Direction(str, Enum):
    LEFT = "left"
    RIGHT = "right"
    UP = "up"
    DOWN = "down"


class RelativeTo(str, Enum):
    AREA = "area"
    PARSER = "parser"
    START = "start"


class Content(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def content(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        pass


@dataclass(config=PydanticConfig)
class Value(JsonEncorder, Content):
    value: typing.Any

    def content(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        return self.value

    def __next__(self) -> "Value":
        return Value(self.value)

    def __getitem__(self, item: int) -> "Value":
        if isinstance(item, slice):
            raise NotImplementedError("Slicing is not supported use itertools and generators")
        if not isinstance(item, int):
            raise ValueError("Only int are allowed as keys")

        return next(self)


@dataclass(config=PydanticConfig)
class SheetAttr(JsonEncorder, Content):
    sheet_attr: str

    def content(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        return getattr(sheet, self.sheet_attr)

    def __iter__(self):
        return self

    def __next__(self) -> "SheetAttr":
        return SheetAttr(self.sheet_attr)

    def __getitem__(self, item: int) -> "SheetAttr":
        if isinstance(item, slice):
            raise NotImplementedError("Slicing is not supported use itertools and generators")
        if not isinstance(item, int):
            raise ValueError("Only int are allowed as keys")

        return next(self)


@dataclass(config=PydanticConfig)
class Coord(JsonEncorder, Content):
    row: int
    col: int
    row_relative_to: RelativeTo = RelativeTo.AREA

    def row_absolute(
        self,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        if self.row_relative_to == RelativeTo.AREA:
            if area_row_offset is None:
                raise RuntimeError("Cord with row_relative_to area need to have area offset set")
            return self.row + area_row_offset
        elif self.row_relative_to == RelativeTo.PARSER:
            if parser_row_offset is None:
                raise RuntimeError(
                    "Cord with row_relative_to parser need to have parser offset set"
                )
            return self.row + parser_row_offset
        elif self.row_relative_to == RelativeTo.START:
            return self.row

    def content(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        row_absolute = self.row_absolute(parser_row_offset, area_row_offset)
        try:
            return sheet[row_absolute][self.col]
        except IndexError as e:
            raise TableException(
                row=row_absolute,
                col=self.col,
                sheet=sheet.sheet_idx,
                reason="out-of-bounds",
                action=TableException.Action.STOP,
            ) from e

    def __iter__(self):
        return self

    def __next__(self) -> "Coord":
        return Coord(self.row, self.col, self.row_relative_to)

    def __getitem__(self, item: int) -> "Coord":
        if isinstance(item, slice):
            raise NotImplementedError("Slicing is not supported use itertools and generators")
        if not isinstance(item, int):
            raise ValueError("Only int are allowed as keys")

        return next(self)

    def __add__(self, other) -> "Coord":
        return Coord(self.row + other.row, self.col + other.col, self.row_relative_to)


@dataclass(config=PydanticConfig)
class CoordRange(JsonEncorder, Content):
    coord: Coord
    direction: Direction
    max_count: typing.Optional[int] = None

    def content(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        return self[self.distance].content(sheet, parser_row_offset, area_row_offset)

    def row_absolute(
        self,
        parser_row_offset: typing.Optional[int] = None,
        area_row_offset: typing.Optional[int] = None,
    ):
        return self[self.distance].row_absolute(parser_row_offset, area_row_offset)

    def __contains__(self, item: Coord) -> bool:
        if not isinstance(item, Coord):
            raise TypeError(f"is '{type(item)}' not Coord")

        if self.direction == Direction.LEFT:
            res = item.row == self.coord.row and item.col <= self.coord.col
            if self.max_count is not None:
                res = res and abs(item.col - self.coord.col) < self.max_count
            return res

        elif self.direction == Direction.RIGHT:
            res = item.row == self.coord.row and item.col >= self.coord.col
            if self.max_count is not None:
                res = res and abs(item.col - self.coord.col) < self.max_count
            return res

        elif self.direction == Direction.UP:
            res = item.col == self.coord.col and item.row <= self.coord.row
            if self.max_count is not None:
                res = res and abs(item.row - self.coord.row) < self.max_count
            return res

        elif self.direction == Direction.DOWN:
            res = item.col == self.coord.col and item.row >= self.coord.row
            if self.max_count is not None:
                res = res and abs(item.row - self.coord.row) < self.max_count
            return res

        raise NotImplementedError()

    def __iter__(self):
        self.distance = 0
        return self

    def __next__(self) -> "Coord":
        col = self.coord.col
        row = self.coord.row
        if self.direction == Direction.LEFT:
            if self.coord.col - self.distance < 0:
                raise StopIteration
            col -= self.distance
        elif self.direction == Direction.RIGHT:
            col += self.distance
        elif self.direction == Direction.UP:
            if self.coord.row - self.distance < 0:
                raise StopIteration
            row -= self.distance
        elif self.direction == Direction.DOWN:
            row += self.distance
        self.distance += 1

        if self.max_count is not None:
            if self.max_count < self.distance:
                raise StopIteration

        return Coord(row, col, self.coord.row_relative_to)

    def __getitem__(self, item: int) -> Coord:
        if isinstance(item, slice):
            raise NotImplementedError("Slicing is not supported use itertools and generators")
        if not isinstance(item, int):
            raise ValueError("Only int are allowed as keys")

        self.distance = item

        if self.max_count is not None:
            if self.max_count < self.distance:
                raise IndexError(f"{item} is not in range of {self}")

        try:
            return next(self)
        except StopIteration:
            raise IndexError(f"{item} is not in range of {self}")

    def skip(self, count: int) -> "CoordRange":
        return CoordRange(self[count], self.direction)



import abc
import typing
from dataclasses import field
from enum import Enum

from celus_nibbler.aggregator import BaseAggregator, NoAggregator, SameAggregator
from celus_nibbler.conditions import Condition
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.base import BaseArea, BaseParser
from celus_nibbler.sources import SpecialExtraction
from celus_nibbler.validators import validators

ValidatorChoices = Enum("ValidatorChoices", [(e.name, e.name) for e in validators])


class BaseParserDefinition(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def make_parser(self) -> typing.Type[BaseParser]:
        pass


class BaseNonCounterParserDefinition(BaseParserDefinition):
    version: int
    kind: str
    parser_name: str
    data_format: DataFormatDefinition
    group: str = "non_counter"
    heuristics: typing.Optional[Condition] = None
    metrics_to_skip: typing.List[str] = []
    available_metrics: typing.Optional[typing.List[str]] = None
    on_metric_check_failed: TableException.Action = TableException.Action.SKIP
    titles_to_skip: typing.List[str] = []
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = {}
    metric_aliases: typing.List[typing.Tuple[str, str]] = []
    metric_value_extraction_overrides: typing.Dict[str, SpecialExtraction] = field(
        default_factory=lambda: {}
    )
    dimension_aliases: typing.List[typing.Tuple[str, str]] = []
    dimensions_validators: typing.Dict[str, ValidatorChoices] = {}


class BaseAreaDefinition(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def kind(self) -> str:
        pass

    @abc.abstractmethod
    def make_area(self) -> typing.Type[BaseArea]:
        pass

    def make_aggregator(self) -> BaseAggregator:
        return (
            SameAggregator() if getattr(self, "aggregate_same_records", False) else NoAggregator()
        )



import csv
import io
import itertools
import logging
import pathlib
import tempfile
from abc import ABCMeta, abstractmethod
from collections import deque
from functools import lru_cache
from io import BufferedIOBase, BytesIO, RawIOBase, StringIO, TextIOBase, TextIOWrapper
from typing import IO, Any, Dict, List, Optional, Sequence, Union

import openpyxl
from celus_nigiri.counter5 import Counter5ReportBase
from celus_nigiri.counter51 import Counter51PRReport
from celus_nigiri.csv_detect import detect_csv_dialect, detect_file_encoding
from celus_nigiri.exceptions import SushiException

from .errors import XlsError

logger = logging.getLogger(__name__)


class _TextIOWrapperNoClose(TextIOWrapper):
    """Wrapper around TextIOWrapper which can't be closed

    Usually TextIOWrapper closes the opened file when it is discarded in `__del__` method.
    This behaviour is not desired, because the underlying file can't be reused afterwards

    This class simply supresses such behaviour
    """

    def __del__(self):
        # Override __del__ method of parent and do nothing
        pass


class SheetReader(metaclass=ABCMeta):
    @property
    @abstractmethod
    def sheet_idx(self) -> int:
        pass

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        pass

    @property
    @abstractmethod
    def extra(self) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def __getitem__(self, item):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __next__(self):
        pass

    @abstractmethod
    def close(self):
        pass

    def dict_reader(self) -> "DictReader":
        return DictReader(SheetReaderWithLineNum(self))


class CsvSheetReader(SheetReader):
    """
    Class representing a single table
    """

    WINDOW_SIZE = 1000  # number of lines to be cached
    sheet_idx = 0
    name = None
    extra = None

    def __init__(
        self,
        sheet_idx: int,
        name: Optional[str],
        file: IO[str],
        window_size: int = WINDOW_SIZE,
        dialect: Optional[str] = None,
    ):
        self.name = name
        self.sheet_idx = sheet_idx
        self.file = file

        self.dialect = dialect or detect_csv_dialect(file)
        self.csv_reader = csv.reader(file, self.dialect)
        # Load basic window
        self.window_start = 0
        self.window_size = window_size
        self.window: Optional[deque[List[str]]] = None
        self.update_window(0)

    def update_window(self, window_start: int):
        if self.window is not None and self.window_start <= window_start:
            # moving forward
            if self.window_start <= window_start - self.window_size:
                # not overlapping
                start = window_start - self.window_start - self.window_size
                stop = window_start - self.window_start
                self.window = deque(
                    itertools.islice(
                        self.csv_reader,
                        start,
                        stop,
                    ),
                    self.window_size,
                )
                self.window_start = window_start
            else:
                # overlapping
                while self.window_start < window_start:
                    self.inc_window()
            return

        # moving backward or at start
        self.file.seek(0)
        self.csv_reader = csv.reader(self.file, self.dialect)
        self.window_start = window_start
        self.window = deque(
            itertools.islice(
                self.csv_reader, self.window_start, self.window_start + self.window_size
            ),
            self.window_size,
        )

    def inc_window(self):
        try:
            row = next(self.csv_reader)
            self.window_start += 1
            self.window.append(row)
        except StopIteration:
            if len(self.window):
                self.window_start += 1
                self.window.popleft()

    @lru_cache(WINDOW_SIZE * 2)  # cache lines to avoid rewinding while reading the header
    def __getitem__(self, item) -> Sequence[str]:
        if isinstance(item, slice):
            raise NotImplementedError("Slicing is not supported use itertools and generators")

        # in current window
        if self.window_start <= item < (self.window_start + self.window_size):
            if self.window_start + len(self.window) < item:
                raise IndexError(f"{item} is out of range")
            return self.window[item - self.window_start]

        # Set window
        self.update_window(item)
        if len(self.window) < 1:
            raise IndexError(f"{item} is out of range")
        return self.window[0]

    def __next__(self):
        if len(self.window) > 0:
            row = self.window[0]
            self.inc_window()
            return row
        else:
            raise StopIteration

    @lru_cache
    def __len__(self):
        res = 0
        self.update_window(0)
        while self.window:
            res += len(self.window)
            self.update_window(self.window_start + self.window_size)
        return res

    def close(self):
        self.file.close()


class JsonCounter5SheetReader(SheetReader):
    WINDOW_SIZE = 1000  # number of lines to be cached
    sheet_idx = 0
    name = None
    extra = None

    def reset(self):
        try:
            # Try to parse counter 5 format first
            report = Counter5ReportBase()

            self.file.seek(0)

            header, items = report.fd_to_dicts(self.file)
        except SushiException:
            # Perhaps it is counter 51
            report = Counter51PRReport()

            self.file.seek(0)
            header, items = report.fd_to_dicts(self.file)

        self.items = items
        self.extra = header

    def __init__(
        self,
        file: IO[bytes],
        window_size: int = WINDOW_SIZE,
    ):
        self.file = file

        self.window_start = 0
        self.window_size = window_size
        self.update_window(0)

    def update_window(self, window_start: int):
        self.reset()
        self.window_start = window_start
        self.window = deque(
            itertools.islice(self.items, self.window_start, self.window_start + self.window_size),
            self.window_size,
        )

    def inc_window(self):
        try:
            next_dict = next(self.items)
            self.window_start += 1
            self.window.append(next_dict)
        except StopIteration:
            if len(self.window):
                self.window_start += 1
                self.window.popleft()

    @lru_cache(WINDOW_SIZE * 2)  # cache lines to avoid rewinding while reading the header
    def __getitem__(self, item) -> Sequence[dict]:
        if isinstance(item, slice):
            raise NotImplementedError("Slicing is not supported use itertools and generators")

        # in current window
        if self.window_start <= item < (self.window_start + self.window_size):
            if self.window_start + len(self.window) < item:
                raise IndexError(f"{item} is out of range")
            return self.window[item - self.window_start]

        # Set window
        self.update_window(item)
        if len(self.window) < 1:
            raise IndexError(f"{item} is out of range")
        return self.window[0]

    def __next__(self):
        if len(self.window) > 0:
            item_dict = self.window[0]
            self.inc_window()
            return item_dict
        else:
            raise StopIteration

    def __len__(self):
        res = 0
        while self.window:
            res += len(self.window)
            self.update_window(self.window_start + self.window_size)
        return res

    def close(self):
        self.file.close()

    def dict_reader(self):
        raise NotImplementedError()


class TableReader(metaclass=ABCMeta):
    """
    Abstract reader for tabular data - defines the API to be used by parsers when reading input data
    """

    @abstractmethod
    def __getitem__(self, item) -> SheetReader:
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()


class CsvReader(TableReader):
    """
    Reads CSV file in stream mode
    """

    def __init__(self, source: Union[IO, bytes, str, pathlib.Path]):
        file: IO[str]
        if isinstance(source, bytes):
            encoding = detect_file_encoding(io.BytesIO(source))
            logger.debug("Encoding '%s' was found for csv data", encoding)
            file = StringIO(source.decode(encoding or "utf8"))
        elif isinstance(source, (str, pathlib.Path)):
            source = pathlib.Path(source)  # make user that source is Path
            with source.open("rb") as f:
                encoding = detect_file_encoding(f)
            logger.debug("Encoding '%s' was found for csv file", encoding)
            file = open(source, "r", encoding=encoding)
        elif isinstance(source, (RawIOBase, BufferedIOBase)):
            encoding = detect_file_encoding(source)
            logger.debug("Encoding '%s' was found for csv data", encoding)
            file = _TextIOWrapperNoClose(source, encoding=encoding)
        elif isinstance(source, TextIOBase):
            # Opened as a text file => don't try to detect encoding
            file = source
        else:
            raise NotImplementedError()

        self.sheets = [CsvSheetReader(0, None, file)]

    def __getitem__(self, item) -> SheetReader:
        return self.sheets[item]

    def __iter__(self):
        return self.sheets.__iter__()


class XlsxReader(TableReader):
    """
    Reads XLSX file in stream mode (TODO verify this)
    """

    def __init__(self, source: Union[str, pathlib.Path, RawIOBase, BufferedIOBase]):
        workbook = openpyxl.load_workbook(source, read_only=True, data_only=True, keep_links=False)
        self.sheets = []

        # Store each sheet as temporary CSV file
        for idx, sheet in enumerate(workbook.worksheets):
            # For some reason in it necessary to reset dimension for some files
            # which display that only a single cell is present in the data
            if sheet.calculate_dimension(force=True) == "A1:A1":
                sheet.reset_dimensions()

            # write data to csv
            f = tempfile.TemporaryFile("w+")
            # unix dialect escapes all by default
            dialect = csv.get_dialect("unix")
            writer = csv.writer(f, dialect=dialect)
            row_length = 0
            for row in sheet.rows:
                # Make sure that length of the row is extending
                current_length = len(row)
                row_length = max(row_length, current_length)
                extra_cells = [""] * (row_length - current_length)

                writer.writerow([cell.value for cell in row] + extra_cells)
            f.seek(0)
            self.sheets.append(CsvSheetReader(idx, workbook.sheetnames[idx], f, dialect="unix"))

        workbook.close()

    def __getitem__(self, item) -> SheetReader:
        return self.sheets[item]

    def __iter__(self):
        return self.sheets.__iter__()


class JsonCounter5Reader(TableReader):
    """Reads JSON file (in Counter 5 format) in stream mode"""

    def __init__(self, source: Union[str, pathlib.Path, bytes]):
        file: IO[bytes]
        if isinstance(source, bytes):
            file = BytesIO(source)
        elif isinstance(source, (str, pathlib.Path)):
            source = pathlib.Path(source)  # make user that source is Path
            file = open(source, "rb")
        else:
            raise ValueError("source")

        self.sheets = [JsonCounter5SheetReader(file)]

    def __getitem__(self, item) -> SheetReader:
        return self.sheets[item]

    def __iter__(self):
        return self.sheets.__iter__()


try:
    import xlrd
except ImportError:
    XlsReader = None
else:

    class XlsReader(TableReader):  # noqa
        """
        Reads XLS file it probably loads entire file into memory
        """

        def __init__(self, source: Union[str, pathlib.Path, RawIOBase, BufferedIOBase]):
            try:
                if isinstance(source, (RawIOBase, BufferedIOBase)):
                    workbook = xlrd.open_workbook(file_contents=source.read())
                else:
                    workbook = xlrd.open_workbook(filename=source)

                self.sheets = []

                # Store each sheet as temporary CSV file
                for idx in range(workbook.nsheets):
                    sheet = workbook.sheet_by_index(idx)

                    # write data to csv
                    f = tempfile.TemporaryFile("w+")
                    # unix dialect escapes all by default
                    dialect = csv.get_dialect("unix")
                    writer = csv.writer(f, dialect=dialect)
                    row_length = sheet.ncols
                    for rx in range(sheet.nrows):
                        row = sheet.row(rx)

                        # Make sure that length of the row is extending
                        current_length = len(row)
                        extra_cells = [""] * (row_length - current_length)

                        writer.writerow([self._cell_to_str(cell) for cell in row] + extra_cells)
                    f.seek(0)

                    self.sheets.append(CsvSheetReader(idx, sheet.name, f, dialect="unix"))
                    workbook.unload_sheet(idx)

                workbook.release_resources()
            except xlrd.compdoc.CompDocError as e:
                raise XlsError(e) from e

        def _cell_to_str(self, cell: xlrd.sheet.Cell) -> str:
            if cell.ctype in [xlrd.XL_CELL_EMPTY, xlrd.XL_CELL_BLANK, xlrd.XL_CELL_ERROR]:
                return ""

            elif cell.ctype == xlrd.XL_CELL_TEXT:
                return cell.value
            elif cell.ctype == xlrd.XL_CELL_BOOLEAN:
                return str(bool(cell.value))
            elif cell.ctype == xlrd.XL_CELL_NUMBER:
                # value is float
                return str(cell.value)
            elif cell.ctype == xlrd.XL_CELL_DATE:
                # value is float
                return xlrd.xldate.xldate_as_datetime(cell.value, 0).isoformat(sep=" ")

            raise NotImplementedError()

        def __getitem__(self, item) -> SheetReader:
            return self.sheets[item]

        def __iter__(self):
            return self.sheets.__iter__()


class SheetReaderWithLineNum:
    def __init__(self, sheet_reader: SheetReader):
        self.sheet_reader = sheet_reader
        self.line_num = 0

    def __next__(self):
        next_element = next(self.sheet_reader)
        self.line_num += 1
        return next_element


class DictReader(csv.DictReader):
    """Try to wrap csv.DictReader arond SheetReader"""

    def __init__(self, sheet_reader: "SheetReaderWithLineNum"):
        self.reader = sheet_reader
        self._fieldnames = None
        self.restkey = None
        self.restval = None
        self.dialect = "excel"
        self.line_num = 0



import typing
from dataclasses import field

from pydantic.dataclasses import dataclass

from celus_nibbler.conditions import Condition
from celus_nibbler.data_headers import DataFormatDefinition, DataHeaders
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.base import BaseArea, BaseTabularParser
from celus_nibbler.parsers.non_counter.generic import BaseGenericArea
from celus_nibbler.sources import (
    AuthorsSource,
    DateSource,
    DimensionSource,
    ItemIdSource,
    ItemSource,
    MetricSource,
    OrganizationSource,
    PublicationDateSource,
    SpecialExtraction,
    TitleIdSource,
    TitleSource,
)
from celus_nibbler.utils import JsonEncorder, PydanticConfig
from celus_nibbler.validators import validators

from .base import BaseAreaDefinition, BaseNonCounterParserDefinition, ValidatorChoices


@dataclass(config=PydanticConfig)
class GenericAreaDefinition(JsonEncorder, BaseAreaDefinition):
    data_headers: DataHeaders
    metrics: typing.Optional[MetricSource] = None
    dates: typing.Optional[DateSource] = None
    titles: typing.Optional[TitleSource] = None
    title_ids: typing.List[TitleIdSource] = field(default_factory=lambda: [])
    items: typing.Optional[ItemSource] = None
    item_ids: typing.List[ItemIdSource] = field(default_factory=lambda: [])
    item_publication_date: typing.Optional[PublicationDateSource] = None
    item_authors: typing.Optional[AuthorsSource] = None
    dimensions: typing.List[DimensionSource] = field(default_factory=lambda: [])
    organizations: typing.Optional[OrganizationSource] = None
    aggregate_same_records: bool = False
    max_areas_generated: typing.Optional[int] = 1
    min_valid_areas: typing.Optional[int] = 1

    kind: typing.Literal["non_counter.generic"] = "non_counter.generic"

    def make_area(self) -> typing.Type[BaseArea]:
        headers = self.data_headers
        titles = self.titles
        title_ids = self.title_ids
        items = self.items
        item_ids = self.item_ids
        item_publication_date = self.item_publication_date
        item_authors = self.item_authors
        dimensions = self.dimensions
        metrics = self.metrics
        organizations = self.organizations
        dates = self.dates
        self_max_areas_generated = self.max_areas_generated
        self_min_valid_areas = self.min_valid_areas

        class Area(BaseGenericArea):
            data_headers = headers
            date_source = dates
            organization_source = organizations
            title_source = titles
            item_source = items
            item_publication_date_source = item_publication_date
            item_authors_source = item_authors
            metric_source = metrics
            aggregator = self.make_aggregator()
            max_areas_generated = self_max_areas_generated
            min_valid_areas = self_min_valid_areas

            @property
            def title_ids_sources(self) -> typing.Dict[str, TitleIdSource]:
                return {e.name: e for e in title_ids}

            @property
            def item_ids_sources(self) -> typing.Dict[str, ItemIdSource]:
                return {e.name: e for e in item_ids}

            @property
            def dimensions_sources(self) -> typing.Dict[str, DimensionSource]:
                return {e.name: e for e in dimensions}

            @property
            def dimensions(self) -> typing.List[str]:
                return list([e.name for e in dimensions])

        return Area


@dataclass(config=PydanticConfig)
class GenericDefinition(JsonEncorder, BaseNonCounterParserDefinition):
    parser_name: str
    data_format: DataFormatDefinition

    areas: typing.List[GenericAreaDefinition] = field(default_factory=lambda: [])
    platforms: typing.List[str] = field(default_factory=lambda: [])
    metrics_to_skip: typing.List[str] = field(default_factory=lambda: [])
    available_metrics: typing.Optional[typing.List[str]] = None
    on_metric_check_failed: TableException.Action = TableException.Action.SKIP
    titles_to_skip: typing.List[str] = field(default_factory=lambda: [])
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = field(default_factory=lambda: {})
    dimensions_validators: typing.Dict[str, ValidatorChoices] = field(default_factory=lambda: {})
    metric_aliases: typing.List[typing.Tuple[str, str]] = field(default_factory=lambda: [])
    metric_value_extraction_overrides: typing.Dict[str, SpecialExtraction] = field(
        default_factory=lambda: {}
    )
    dimension_aliases: typing.List[typing.Tuple[str, str]] = field(default_factory=lambda: [])
    heuristics: typing.Optional[Condition] = None
    possible_row_offsets: typing.List[int] = field(default_factory=lambda: [0])

    kind: typing.Literal["non_counter.generic"] = "non_counter.generic"
    version: typing.Literal[1] = 1

    def make_parser(self):
        validators_map = {e.name: e for e in validators}
        _dimensions_validators = {
            dimension_name: validators_map[validator_name.value]
            for dimension_name, validator_name in self.dimensions_validators.items()
        }

        class Parser(BaseTabularParser):
            _definition = self

            data_format = _definition.data_format
            platforms = _definition.platforms
            name = (
                f"dynamic.{_definition.group}.{_definition.data_format.name}"
                f".{_definition.parser_name}"
            )

            metrics_to_skip = _definition.metrics_to_skip
            available_metrics = _definition.available_metrics
            on_metric_check_failed = _definition.on_metric_check_failed
            titles_to_skip = _definition.titles_to_skip
            dimensions_to_skip = _definition.dimensions_to_skip
            dimensions_validators = _dimensions_validators

            metric_aliases = dict(_definition.metric_aliases)
            metric_value_extraction_overrides = dict(_definition.metric_value_extraction_overrides)
            dimension_aliases = dict(_definition.dimension_aliases)

            heuristics = _definition.heuristics
            possible_row_offsets = _definition.possible_row_offsets

            areas = [e.make_area() for e in _definition.areas]

        return Parser



import abc
import copy
import tempfile
import typing
from collections import deque
from datetime import date

from celus_nigiri import CounterRecord
from diskcache import Cache

from celus_nibbler.errors import NegativeValueInOutput, SameRecordsInOutput


class BaseAggregator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        pass

    def __or__(self, other):
        return PippedAggregator(self, other)


class NoAggregator(BaseAggregator):
    """Doesn't aggregate anything just passes the data"""

    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        yield from records


class SameAggregator(BaseAggregator):
    """Aggregates the same records - sums values"""

    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as db, db.transact():
                # Read all records
                for record in records:
                    # Derive key
                    new_record = copy.deepcopy(record)
                    new_record.value = -1
                    key = new_record.as_csv()

                    # Aggregate
                    if existing_record := db.get(key):
                        existing_record.value += record.value
                        db[key] = existing_record
                    else:
                        db[key] = record

                for key in db.iterkeys():
                    yield db[key]


class CheckConflictingRecordsAggregator(BaseAggregator):
    """
    Checks whether last n records doesn't contain same conflicting records e.g.

    Title1,Dim1,Metric1,2021-01-01,2021-01-31,999
    Title2,Dim1,Metric1,2021-01-01,2021-01-31,888
    Title1,Dim1,Metric1,2021-01-01,2021-01-31,999

    is wrong, because it contains same Title, Dimensions and metric for 2020-01 twice

    """

    def __init__(self, size: int = 100):
        self.hash_buffer: typing.Deque[int, int] = deque(maxlen=size)

    @staticmethod
    def make_record_hash(record: CounterRecord):
        return hash(tuple(e for e in record.as_csv()[:-1]))  # skip value (last in csv)

    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        idx = 0
        for record in records:
            hsh = self.make_record_hash(record)
            for e in self.hash_buffer:
                if e[0] == hsh:
                    raise SameRecordsInOutput(e[1], idx, record)
            self.hash_buffer.append((hsh, idx))
            yield record
            idx += 1


class CounterOrdering(BaseAggregator):
    """
    Extracts same records with different dates together e.g.

    Title1,Dim1,Metric1,2021-01-01,2021-01-31,999
    Title2,Dim1,Metric1,2021-01-01,2021-01-31,888
    Title1,Dim1,Metric1,2021-02-01,2021-02-28,777
    Title2,Dim1,Metric1,2021-02-01,2021-02-28,666

    ->

    Title1,Dim1,Metric1,2021-01-01,2021-01-31,999
    Title1,Dim1,Metric1,2021-02-01,2021-02-28,777
    Title2,Dim1,Metric1,2021-01-01,2021-01-31,888
    Title2,Dim1,Metric1,2021-02-01,2021-02-28,666
    """

    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        with tempfile.TemporaryDirectory() as tmpdir:
            with Cache(tmpdir) as db, db.transact():
                # Read all records
                for record in records:
                    # Derive key
                    new_record = copy.deepcopy(record)
                    new_record.value = -1
                    # Date aggregation -> fix date in key
                    new_record.start = date(2020, 1, 1)
                    new_record.end = date(2020, 1, 31)
                    key = new_record.as_csv()

                    # Aggregate
                    record_dict: typing.Dict[date, CounterRecord] = db.get(key, {})
                    if rec := record_dict.get(record.start):
                        rec.value += record.value
                    else:
                        record_dict[record.start] = record
                    db[key] = record_dict

                for db_key in db.iterkeys():
                    record_dict = db[db_key]
                    yield from [record_dict[key] for key in sorted(record_dict.keys())]


class CheckNonNegativeValues(BaseAggregator):
    """Checks whether outout contains negative values and raises and error if so"""

    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        idx = 0
        for record in records:
            if record.value < 0:
                raise NegativeValueInOutput(idx, record)
            yield record
            idx += 1


class PippedAggregator(BaseAggregator):
    """Pipes output from one aggregator to another aggregator"""

    def __init__(self, a1: BaseAggregator, a2: BaseAggregator):
        self.a1 = a1
        self.a2 = a2

    def aggregate(
        self, records: typing.Generator[CounterRecord, None, None]
    ) -> typing.Generator[CounterRecord, None, None]:
        return self.a2.aggregate(self.a1.aggregate(records))



import datetime
import inspect
import re
from functools import lru_cache
from typing import Any, List, Optional, Tuple, Type, Union

from celus_nigiri.record import Author
from dateutil import parser as datetimes_parser
from isbnlib import get_isbnlike, is_isbn10, is_isbn13
from pydantic import NonNegativeFloat, NonNegativeInt, field_validator
from pydantic.dataclasses import dataclass as pydantic_dataclass

from .utils import COMMON_DATE_FORMATS, PydanticConfig

issn_matcher = re.compile(r"(\d{4})-?(\d{3}[\dXx])")
issn_number_matcher = re.compile(r"^\d{0,7}[\dXx]$")


@pydantic_dataclass(config=PydanticConfig)
class BaseValueModel:
    model_config = PydanticConfig
    name = "base"

    value: Any


def non_empty(name: str) -> str:
    if not name:
        raise ValueError("cant-be-empty")
    return name


def stripped(name: Optional[str]) -> Optional[str]:
    if name and isinstance(name, str):
        return name.strip()
    return name


def not_none(value: Any) -> Any:
    if value is None:
        raise ValueError("cant-be-none")
    return value


def issn(issn: str) -> str:
    return issn.strip() or ""


def issn_strict(issn: str) -> str:
    clean = "".join(issn.split())  # remove all whitespace

    if m := issn_matcher.search(clean):
        # upper() because 'X' can be also lowercase
        return m.group(1) + "-" + m.group(2).upper()

    # sometimes the leading zeros are missing, so we add them
    if issn_number_matcher.match(clean):
        clean = (8 - len(clean)) * "0" + clean
        return clean[:4] + "-" + clean[4:].upper()

    raise ValueError(f'Invalid ISSN: "{issn}"')


def date_aligned(value: datetime.datetime) -> datetime.datetime:
    if value.day != 1:
        raise ValueError("date-not-aligned")
    return value


def remove_multiple_spaces(value: Optional[str]) -> str:
    return value and re.sub(r"\s+", " ", value)


@pydantic_dataclass(config=PydanticConfig)
class Value(BaseValueModel):
    name = "value"
    value: Union[NonNegativeInt, NonNegativeFloat]

    _stripped_value = field_validator("value", mode="before")(stripped)

    @field_validator("value")
    def non_negative(cls, value: Union[NonNegativeInt, NonNegativeFloat]) -> int:
        return round(value)


@pydantic_dataclass(config=PydanticConfig)
class ValueNegative(BaseValueModel):
    name = "value_negative"
    value: Union[int, float]

    @field_validator("value")
    def non_negative(cls, value: Union[int, float]) -> Union[int, float]:
        return round(value)


@lru_cache
def gen_default_validator(
    orig_validator: Type[BaseValueModel], default_value: Any, blank_values: Tuple[Any]
) -> Type[BaseValueModel]:
    @pydantic_dataclass(config=PydanticConfig)
    class Validator(BaseValueModel):
        name = f"{orig_validator.name}_default"
        value: Any

        @field_validator("value")
        def default(cls, value: Any) -> Any:
            if value in blank_values:
                return default_value
            else:
                return orig_validator(value=value).value

    return Validator


@pydantic_dataclass(config=PydanticConfig)
class CommaSeparatedNumberValidator(BaseValueModel):
    name = "comma_separeted_number"
    value: Any

    @field_validator("value", mode="before")
    def comma_separeted_number(cls, value: str) -> str:
        return Value(value=value.replace(",", "")).value


@pydantic_dataclass(config=PydanticConfig)
class MinutesToSecondsValidator(BaseValueModel):
    name = "minutes_to_seconds"
    value: Union[NonNegativeInt, NonNegativeFloat]

    @field_validator("value", mode="after")
    def minutes_to_seconds(cls, value: Union[float, int]) -> int:
        return round(value * 60)


@pydantic_dataclass(config=PydanticConfig)
class HoursToSecondsValidator(BaseValueModel):
    name = "hours_to_seconds"
    value: Union[NonNegativeInt, NonNegativeFloat]

    @field_validator("value", mode="after")
    def hours_to_seconds(cls, value: Union[float, int]) -> int:
        return round(value * 60 * 60)


@pydantic_dataclass(config=PydanticConfig)
class DurationToSecondsValidator(BaseValueModel):
    """Converts HH:MM:SS to format to seconds"""

    name = "duration_to_seconds"
    value: Union[NonNegativeInt, NonNegativeFloat]

    @field_validator("value", mode="before")
    def duration_to_seconds(cls, value: str) -> int:
        try:
            hours, minutes, seconds = value.split(":")
            return 60 * 60 * int(hours) + 60 * int(minutes) + int(seconds)
        except ValueError:
            raise ValueError("cant-parse-duration")


@pydantic_dataclass(config=PydanticConfig)
class Organization(BaseValueModel):
    name = "organization"
    value: str

    _not_none_organization = field_validator("value")(not_none)
    _stripped_organization = field_validator("value")(stripped)
    _non_empty_organization = field_validator("value")(non_empty)
    _remove_multiple_spaces = field_validator("value")(remove_multiple_spaces)


@pydantic_dataclass(config=PydanticConfig)
class Platform(BaseValueModel):
    name = "platform"
    value: str

    _stripped_platform = field_validator("value")(stripped)
    _non_empty_platform = field_validator("value")(non_empty)


@pydantic_dataclass(config=PydanticConfig)
class Dimension(BaseValueModel):
    name = "dimension"
    value: str

    _stripped_dimension = field_validator("value")(stripped)
    _remove_multiple_spaces = field_validator("value")(remove_multiple_spaces)


@pydantic_dataclass(config=PydanticConfig)
class Metric(BaseValueModel):
    name = "metric"
    value: str

    _not_none_metic = field_validator("value")(not_none)
    _stripped_metric = field_validator("value")(stripped)
    _non_empty_metric = field_validator("value")(non_empty)
    _remove_multiple_spaces = field_validator("value")(remove_multiple_spaces)

    @field_validator("value")
    def not_digit(cls, metric: str) -> str:
        if metric.isdigit():
            raise ValueError("cant-be-digit")
        return metric


@pydantic_dataclass(config=PydanticConfig)
class Title(BaseValueModel):
    name = "title"
    value: Optional[str]

    _stripped_title = field_validator("value")(stripped)
    _remove_multiple_spaces = field_validator("value")(remove_multiple_spaces)


parserinfo_us = datetimes_parser.parserinfo(dayfirst=False)  # prefer US variant
parserinfo_eu = datetimes_parser.parserinfo(dayfirst=True)  # prefer EU variant


@pydantic_dataclass(config=PydanticConfig)
class Date(BaseValueModel):
    name = "date"
    value: datetime.date

    _not_none_date = field_validator("value", mode="before")(not_none)
    _stripped_date = field_validator("value", mode="before")(stripped)
    _non_empty_date = field_validator("value", mode="before")(non_empty)

    @classmethod
    def parserinfo(cls):
        return parserinfo_us

    @classmethod
    def align_date(cls, input: datetime.datetime) -> datetime.datetime:
        return input.replace(day=1)

    @field_validator("value", mode="before")
    def to_datetime(cls, date: str) -> datetime.datetime:
        if not date:
            raise ValueError("no-date-provided")

        # Check for common formats (faster that dateutil)
        for fmt in COMMON_DATE_FORMATS:
            try:
                return cls.align_date(datetime.datetime.strptime(date, fmt))
            except ValueError:
                pass

        # Try to parse date using dateutil for more obscure date formats
        try:
            return cls.align_date(datetimes_parser.parse(date, parserinfo=cls.parserinfo()))
        except datetimes_parser.ParserError:
            raise ValueError("cant-parse-date")


@pydantic_dataclass(config=PydanticConfig)
class DateEU(Date):
    name = "date_eu"

    @classmethod
    def parserinfo(cls):
        return parserinfo_eu


@pydantic_dataclass(config=PydanticConfig)
class DateAligned(Date):
    name = "date_aligned"
    _date_aligned = field_validator("value", mode="after")(date_aligned)

    @classmethod
    def align_date(cls, input: datetime.datetime) -> datetime.datetime:
        # Don't perfrom auto aligment
        return input


@pydantic_dataclass(config=PydanticConfig)
class DateEUAligned(DateEU):
    name = "date_eu_aligned"
    _date_aligned = field_validator("value", mode="after")(date_aligned)

    @classmethod
    def align_date(cls, input: datetime.datetime) -> datetime.datetime:
        # Don't perfrom auto aligment
        return input


@lru_cache
def gen_date_format_validator(pattern: str) -> Type[BaseValueModel]:
    @pydantic_dataclass(config=PydanticConfig)
    class DateFormat(BaseValueModel):
        name = "date_format"
        value: datetime.date

        _not_none_date = field_validator("value", mode="before")(not_none)
        _stripped_date = field_validator("value", mode="before")(stripped)
        _non_empty_date = field_validator("value", mode="before")(non_empty)

        @field_validator("value", mode="before")
        def to_datetime(cls, date: str) -> datetime.date:
            if not date:
                raise ValueError("no-date-provided")
            try:
                parsed = datetime.datetime.strptime(date, pattern)
                return datetime.date(year=parsed.year, month=parsed.month, day=1)
            except ValueError:
                raise ValueError("cant-parse-date")

    return DateFormat


@pydantic_dataclass(config=PydanticConfig)
class DOI(BaseValueModel):
    name = "doi"
    value: str

    @field_validator("value")
    def check_doi(cls, doi: str) -> str:
        return doi.strip() or ""


@pydantic_dataclass(config=PydanticConfig)
class URI(BaseValueModel):
    name = "uri"
    value: str

    @field_validator("value")
    def check_uri(cls, uri: str) -> str:
        return uri.strip() or ""


@pydantic_dataclass(config=PydanticConfig)
class ISBN(BaseValueModel):
    name = "isbn"
    value: str

    @field_validator("value")
    def check_isbn(cls, isbn: str) -> str:
        return isbn.strip() or ""


@pydantic_dataclass(config=PydanticConfig)
class StrictISBN(BaseValueModel):
    name = "isbn_strict"
    value: str

    @field_validator("value")
    def check_isbn(cls, isbn: str) -> str:
        isbns = get_isbnlike(isbn, level="strict")

        if not isbns:
            raise ValueError("isbn-not-valid")

        # return the first isbn the rest is omitted
        isbn = isbns[0]

        # check isbns including checksums
        if not (is_isbn10(isbn) or is_isbn13(isbn)):
            raise ValueError("isbn-not-valid")

        return isbn


@pydantic_dataclass(config=PydanticConfig)
class StrictISBN13(BaseValueModel):
    name = "isbn13"
    value: str

    @field_validator("value")
    def check_isbn(cls, isbn: str) -> str:
        isbns = get_isbnlike(isbn, level="strict")

        if not isbns:
            raise ValueError("isbn13-not-valid")

        # return the first isbn the rest is omitted
        isbn = isbns[0]

        # check isbns including checksums
        if not is_isbn13(isbn):
            raise ValueError("isbn13-not-valid")

        return isbn


@pydantic_dataclass(config=PydanticConfig)
class StrictISBN10(BaseValueModel):
    name = "isbn10"
    value: str

    @field_validator("value")
    def check_isbn(cls, isbn: str) -> str:
        isbns = get_isbnlike(isbn, level="strict")

        if not isbns:
            raise ValueError("isbn10-not-valid")

        # return the first isbn the rest is omitted
        isbn = isbns[0]

        # check isbns including checksums
        if not is_isbn10(isbn):
            raise ValueError("isbn10-not-valid")

        return isbn


@pydantic_dataclass(config=PydanticConfig)
class ISSN(BaseValueModel):
    name = "issn"
    value: str

    _issn_format = field_validator("value")(issn)


@pydantic_dataclass(config=PydanticConfig)
class StrictISSN(BaseValueModel):
    name = "issn_strict"
    value: str

    _issn_format = field_validator("value")(issn_strict)


@pydantic_dataclass(config=PydanticConfig)
class EISSN(BaseValueModel):
    name = "eissn"
    value: str

    _issn_format = field_validator("value")(issn)


@pydantic_dataclass(config=PydanticConfig)
class StrictEISSN(BaseValueModel):
    name = "eissn_strict"
    value: str

    _issn_format = field_validator("value")(issn_strict)


@pydantic_dataclass(config=PydanticConfig)
class ProprietaryID(BaseValueModel):
    name = "proprietary_id"
    value: str


@pydantic_dataclass(config=PydanticConfig)
class AuthorsValidator(BaseValueModel):
    name = "authors"
    value: str

    AUTHOR_REGEX = re.compile(r"^([^\(]+)\s+\(([a-zA-Z]+):([^\)]+)\)$")

    @field_validator("value")
    def authors(cls, authors: str) -> List[Author]:
        res = []
        splitted_authors = [e.strip() for e in authors.split(";")]
        for sa in splitted_authors:
            if match := cls.AUTHOR_REGEX.match(sa):
                name, kind, code = match.groups()
                if kind == "ISNI":
                    res.append(Author(name=name, ISNI=code))
                elif kind == "ORCID":
                    res.append(Author(name=name, ORCID=code))
                else:
                    res.append(Author(name=name))
            else:
                res.append(Author(name=sa))

        return res


@pydantic_dataclass(config=PydanticConfig)
class YOPValidator(BaseValueModel):
    name = "yop"
    value: str

    YOP_REGEX = re.compile(r"^\s*(\d{4})\s*$")

    @field_validator("value")
    def yop(cls, yop: str) -> List[Author]:
        if not cls.YOP_REGEX.match(yop):
            raise ValueError("not-yop")

        return yop


validators = [e for e in locals().values() if inspect.isclass(e) and issubclass(e, BaseValueModel)]



import datetime
import typing
from abc import ABCMeta

from celus_nibbler.parsers.base import BaseHeaderArea

from .base import BaseNonCounterParser


class BaseGenericArea(BaseHeaderArea, metaclass=ABCMeta):
    def get_months(self) -> typing.List[datetime.date]:
        if self.date_source:
            self.find_data_cells(lambda x: x, lambda x, y: None)
            return self._get_months_from_column(0, self.row_offset)
        else:
            # Extract months from header_data
            # populate Area.row_offset
            return self._get_months_from_header()


class BaseGenericParser(BaseNonCounterParser):
    platforms: typing.List[str] = []
    metrics_to_skip: typing.List[str] = []
    titles_to_skip: typing.List[str] = []
    dimensions_to_skip: typing.Dict[str, typing.List[str]] = {}

    heuristics = None



import functools
import itertools
import logging
import pathlib
import typing
from collections import Counter, defaultdict
from datetime import date

from celus_nigiri import CounterRecord
from pydantic import Field
from pydantic.dataclasses import dataclass
from typing_extensions import Annotated

from celus_nibbler.aggregator import CheckConflictingRecordsAggregator, CheckNonNegativeValues
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.errors import (
    MultipleParsersFound,
    NibblerError,
    NoParserForFileTypeFound,
    NoParserForPlatformFound,
    NoParserFound,
    NoParserMatchesHeuristics,
    WrongFileFormatError,
)
from celus_nibbler.parsers import BaseParser, get_parsers
from celus_nibbler.reader import (
    CsvReader,
    JsonCounter5Reader,
    SheetReader,
    TableReader,
    XlsReader,
    XlsxReader,
)
from celus_nibbler.utils import JsonEncorder, PydanticConfig
from celus_nibbler.validators import Platform

logger = logging.getLogger(__name__)


@dataclass(config=PydanticConfig)
class StatUnit(JsonEncorder):
    count: int = 0
    sum: int = 0

    def inc(self, value: int):
        self.count += 1
        self.sum += value

    def __add__(self, other: "StatUnit") -> "StatUnit":
        return StatUnit(
            count=self.count + other.count,
            sum=self.sum + other.sum,
        )

    def __iadd__(self, other: "StatUnit"):
        self.sum += other.sum
        self.count += other.count
        return self


AnnotatedStatUnit = Annotated[StatUnit, Field(default_factory=lambda: StatUnit)]


@dataclass(config=PydanticConfig)
class PoopOrganizationStats(JsonEncorder):
    months: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    metrics: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    titles: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    title_ids: typing.Set[str] = Field(default_factory=lambda: set())
    items: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    item_ids: typing.Set[str] = Field(default_factory=lambda: set())
    dimensions: typing.DefaultDict[
        str,
        Annotated[
            typing.DefaultDict[str, AnnotatedStatUnit],
            Field(default_factory=lambda: defaultdict(StatUnit)),
        ],
    ] = Field(defaultdict(lambda: defaultdict(StatUnit)))
    total: StatUnit = Field(default_factory=lambda: StatUnit())

    def process_record(self, record: CounterRecord):
        self.months[record.start.strftime("%Y-%m")].inc(record.value)
        self.metrics[record.metric or ""].inc(record.value)
        self.titles[record.title or ""].inc(record.value)
        self.items[record.item or ""].inc(record.value)

        for title_id in record.title_ids.keys():
            self.title_ids.add(title_id)

        for item_id in record.item_ids.keys():
            self.item_ids.add(item_id)

        for dimension_name, dimension in record.dimension_data.items():
            self.dimensions[dimension_name][dimension].inc(record.value)

        self.total.inc(record.value)

    def __add__(self, other: "PoopOrganizationStats") -> "PoopOrganizationStats":
        result = PoopOrganizationStats()
        for month, data in list(self.months.items()) + list(other.months.items()):
            result.months[month] += data

        for metric, data in list(self.metrics.items()) + list(other.metrics.items()):
            result.metrics[metric] += data

        for title, data in list(self.titles.items()) + list(other.titles.items()):
            result.titles[title] += data

        for item, data in list(self.items.items()) + list(other.items.items()):
            result.items[item] += data

        for dimension_name, dimensions in list(self.dimensions.items()) + list(
            other.dimensions.items()
        ):
            for dimension, data in dimensions.items():
                result.dimensions[dimension_name][dimension] += data

        result.title_ids = self.title_ids | other.title_ids
        result.item_ids = self.item_ids | other.item_ids

        result.total = self.total + other.total
        return result


@dataclass(config=PydanticConfig)
class PoopStats(JsonEncorder):
    months: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    metrics: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    organizations: typing.DefaultDict[
        str,
        Annotated[PoopOrganizationStats, Field(default_factory=lambda: PoopOrganizationStats())],
    ] = Field(default_factory=lambda: defaultdict(PoopOrganizationStats))
    titles: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    title_ids: typing.Set[str] = Field(default_factory=lambda: set())
    items: typing.DefaultDict[str, AnnotatedStatUnit] = Field(
        default_factory=lambda: defaultdict(StatUnit)
    )
    item_ids: typing.Set[str] = Field(default_factory=lambda: set())
    dimensions: typing.DefaultDict[
        str,
        Annotated[
            typing.DefaultDict[str, AnnotatedStatUnit],
            Field(default_factory=lambda: defaultdict(StatUnit)),
        ],
    ] = Field(defaultdict(lambda: defaultdict(StatUnit)))
    total: StatUnit = Field(default_factory=lambda: StatUnit())

    def __add__(self, other: "PoopStats") -> "PoopStats":
        result = PoopStats()
        for month, data in list(self.months.items()) + list(other.months.items()):
            result.months[month] += data

        for metric, data in list(self.metrics.items()) + list(other.metrics.items()):
            result.metrics[metric] += data

        for organization, data in list(self.organizations.items()) + list(
            other.organizations.items()
        ):
            result.organizations[organization] += data

        for title, data in list(self.titles.items()) + list(other.titles.items()):
            result.titles[title] += data

        for item, data in list(self.items.items()) + list(other.items.items()):
            result.items[item] += data

        for dimension_name, dimensions in list(self.dimensions.items()) + list(
            other.dimensions.items()
        ):
            for dimension, data in dimensions.items():
                result.dimensions[dimension_name][dimension] += data

        result.title_ids = self.title_ids | other.title_ids
        result.item_ids = self.item_ids | other.item_ids

        result.total = self.total + other.total
        return result

    def process_record(self, record: CounterRecord):
        self.months[record.start.strftime("%Y-%m")].inc(record.value)
        self.metrics[record.metric or ""].inc(record.value)
        self.organizations[record.organization or ""].process_record(record)
        self.titles[record.title or ""].inc(record.value)
        self.items[record.item or ""].inc(record.value)

        for title_id in record.title_ids.keys():
            self.title_ids.add(title_id)

        for item_id in record.item_ids.keys():
            self.item_ids.add(item_id)

        for dimension_name, dimension in record.dimension_data.items():
            self.dimensions[dimension_name][dimension].inc(record.value)

        self.total.inc(record.value)


class Poop:
    """nibblonians dark matter"""

    def __init__(self, parser: BaseParser):
        self.parser = parser
        self.current_stats = PoopStats()
        self.area_counter: Counter[int] = Counter()
        self.extras = parser.get_extras()

    @property
    def sheet_idx(self):
        return self.parser.sheet.sheet_idx

    @property
    def data_format(self) -> DataFormatDefinition:
        return self.parser.data_format

    def records_basic(
        self,
        offset: int = 0,
        limit: typing.Optional[int] = None,
        same_check_size: int = 0,
    ) -> typing.Optional[typing.Generator[typing.Tuple[int, CounterRecord], None, None]]:
        if counter_records := self.parser.parse():
            if limit is None:
                return itertools.islice(counter_records, offset, None)
            return itertools.islice(counter_records, offset, offset + limit)

        return None

    def records_with_counter(
        self,
        offset: int = 0,
        limit: typing.Optional[int] = None,
        same_check_size: int = 0,
    ) -> typing.Optional[typing.Generator[CounterRecord, None, None]]:
        self.area_counter = Counter()
        if counter_records := self.records_basic(offset, limit, same_check_size):
            for idx, record in counter_records:
                self.area_counter[idx] += 1
                yield record
        return None

    def records(
        self,
        offset: int = 0,
        limit: typing.Optional[int] = None,
        same_check_size: int = 0,
    ) -> typing.Optional[typing.Generator[CounterRecord, None, None]]:
        if counter_records := self.records_with_counter(offset, limit, same_check_size):
            aggregator = CheckNonNegativeValues()
            if same_check_size:
                aggregator = aggregator | CheckConflictingRecordsAggregator(same_check_size)

            return aggregator.aggregate(counter_records)
        else:
            logger.warning("sheet %s has not been parsed", self.parser.sheet.sheet_idx + 1)
            return None

    def records_with_stats(
        self,
        offset: int = 0,
        limit: typing.Optional[int] = None,
        same_check_size: int = 0,
    ) -> typing.Optional[typing.Generator[CounterRecord, None, None]]:
        self.current_stats = PoopStats()
        if records := self.records(offset, limit, same_check_size):
            for record in records:
                self.current_stats.process_record(record)
                yield record

        return None

    @property
    def metrics(self):
        return self.get_stats().metrics.keys()

    @property
    def dimensions(self):
        return self.get_stats().dimensions.keys()

    @property
    def title_ids(self):
        return self.get_stats().title_ids

    @property
    def item_ids(self):
        return self.get_stats().item_ids

    @property
    def months(self):
        return self.get_stats().months.keys()

    @functools.lru_cache
    def get_stats(self) -> PoopStats:
        """Goes through all records and caluculates stats based on output records"""
        res = PoopStats()
        if records := self.records():
            for record in records:
                res.process_record(record)

        return res

    def get_months(self) -> typing.List[typing.List[date]]:
        """Get months of the sheet (divided into areas)"""
        return self.parser.get_months()

    def __del__(self):
        # Make sure that underlying file is closed
        self.parser.sheet.close()


def findparser(
    sheet: SheetReader,
    platform: str,
    parsers: typing.Optional[typing.List[str]] = None,
    check_platform: bool = True,
    use_heuristics: bool = True,
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> BaseParser:
    parser_classes = [
        (name, parser)
        for name, parser in get_parsers(parsers, dynamic_parsers)
        if not check_platform or parser.check_platform(platform)
    ]

    if len(parser_classes) < 1:
        logger.warning("there is no parser which expects your platform %s", platform)
        raise NoParserForPlatformFound(sheet.sheet_idx)

    parser_classes = [
        (name, parser)
        for name, parser in parser_classes
        if type(sheet) in parser.sheet_reader_classes()
    ]
    if len(parser_classes) < 1:
        logger.warning("no parser found for reader %s", type(sheet))
        raise NoParserForFileTypeFound(sheet.sheet_idx)

    parser_instances = [(name, parser(sheet, platform=platform)) for name, parser in parser_classes]
    parser_instances_filtered = [
        (name, parser)
        for (name, parser) in parser_instances
        if parser.heuristic_check() or not use_heuristics
        # Note that we need to always perform heuristic check in order to calculate offsets
    ]

    if len(parser_instances_filtered) < 1:
        logger.warning("no parser found")
        raise NoParserMatchesHeuristics(
            sheet.sheet_idx,
            parsers_info={name: parser.analyze() for name, parser in parser_instances},
        )

    elif len(parser_instances_filtered) > 1:
        logger.warning("%s more than one parser found", len(parser_classes))
        raise MultipleParsersFound(sheet.sheet_idx, *(e[0] for e in parser_classes))

    name, parser = parser_instances_filtered[0]
    logger.info("Parser used: %s", name)
    return parser


def read_file(file_path: pathlib.Path) -> TableReader:
    if file_path.suffix.lower() in [".csv", ".tsv"]:
        return CsvReader(file_path)
    elif file_path.suffix.lower() == ".xlsx":
        return XlsxReader(file_path)
    elif XlsReader and file_path.suffix.lower() in [".xls", ".xlsb"]:
        return XlsReader(file_path)
    elif file_path.suffix.lower() == ".json":
        return JsonCounter5Reader(file_path)

    raise WrongFileFormatError(file_path, file_path.suffix)


def eat(
    file_path: typing.Union[pathlib.Path, str],
    platform: str,
    parsers: typing.Optional[typing.List[str]] = None,
    check_platform: bool = True,
    use_heuristics: bool = True,
    dynamic_parsers: typing.List[typing.Type[BaseParser]] = [],
) -> typing.List[typing.Union[Poop, NibblerError]]:
    platform = Platform(value=platform).value

    # make sure that file_path is Path instance
    file_path = pathlib.Path(file_path)

    logger.info('Eating file "%s"', file_path)

    reader = read_file(file_path)
    poops = []
    for sheet in reader:
        logger.info("Digesting sheet %d", sheet.sheet_idx)
        try:
            parser = findparser(
                sheet, platform, parsers, check_platform, use_heuristics, dynamic_parsers
            )
            poops.append(Poop(parser))
        except (NoParserFound, MultipleParsersFound) as e:
            logger.warning(
                "parser has not been chosen for sheet %s, the sheet wont be parsed",
                sheet.sheet_idx + 1,
            )
            poops.append(e)
            # Make sure that underlying file is closed
            sheet.close()

    return poops



import re
import typing

from celus_nibbler.conditions import RegexCondition
from celus_nibbler.coordinates import Coord, CoordRange, Direction, RelativeTo
from celus_nibbler.data_headers import DataFormatDefinition
from celus_nibbler.errors import TableException
from celus_nibbler.parsers.base import BaseTabularParser
from celus_nibbler.sources import DimensionSource
from celus_nibbler.validators import BaseValueModel

from . import CounterHeaderArea


class Counter5HeaderArea(CounterHeaderArea):
    HEADER_DATE_COL_START = 3
    METRIC_COLUMN_NAMES = ["Metric_Type", "Metric Type"]
    TITLE_COLUMN_NAMES = ["Title"]
    ORGANIZATION_COLUMN_NAMES = ["Institution_Name", "Institution Name"]


class Counter5ParserAnalyzeMixin:
    def analyze(self) -> typing.List[dict]:
        header = self.get_extras()
        if not header:
            return [
                {
                    "code": "no-header-data-extracted",
                }
            ]
        if "Report_ID" not in header:
            return [
                {
                    "code": "report-id-not-in-header",
                }
            ]
        if header["Report_ID"] != self.data_format.name:
            return [
                {
                    "code": "wrong-report-type",
                    "found": header["Report_ID"],
                    "expected": self.data_format.name,
                }
            ]
        return []


class BaseCounter5Parser(Counter5ParserAnalyzeMixin, BaseTabularParser):
    data_format: DataFormatDefinition
    Area: typing.Type[CounterHeaderArea]

    # TODO perhaps implement some kind of YOP validator
    dimensions_validators: typing.Dict[str, typing.Type[BaseValueModel]] = {}

    @property
    def name(self):
        return f"static.counter5.{self.data_format.name}.Tabular"

    def get_extras(self) -> dict:
        area = self.Area(self.sheet, self.platform)

        # get header row
        try:
            col = area.header_row.coord.col
            row = area.header_row.coord.row
        except TableException:
            return {}

        if row == 0:
            return {}

        res = {}
        for coord in CoordRange(
            Coord(col=col, row=row - 1, row_relative_to=RelativeTo.START), direction=Direction.UP
        ):
            try:
                key = coord.content(self.sheet).strip()
                if not key:
                    continue
            except TableException:
                continue

            try:
                right_coord = Coord(
                    col=coord.col + 1, row=coord.row, row_relative_to=RelativeTo.START
                )
                value = right_coord.content(self.sheet).strip()
            except TableException:
                # missing right cord => consider cell as empty
                value = ""
            res[key] = value

        # Fix 5.0 => 5
        if res.get("Release", "") == "5.0":
            res["Release"] = "5"

        return res


class DR(BaseCounter5Parser):
    data_format = DataFormatDefinition(name="DR")

    titles_to_skip: typing.List[str] = ["Total", "All Databases", ""]

    platforms = ["*"]
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^DR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5(\.0)?$"), Coord(2, 1))
    )

    class Area(Counter5HeaderArea):
        DIMENSION_NAMES_MAP = [
            ("Access_Method", {"Access_Method", "Access Method"}),
            ("Data_Type", {"Data Type", "Data_Type"}),
            ("Platform", {"Platform"}),
            ("Publisher", {"Publisher"}),
        ]
        TITLE_COLUMN_NAMES = ["Title", "Database"]

    areas = [Area]


class PR(BaseCounter5Parser):
    data_format = DataFormatDefinition(name="PR")

    platforms = ["*"]
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^PR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5(\.0)?$"), Coord(2, 1))
    )

    class Area(Counter5HeaderArea):
        title_source = None
        DIMENSION_NAMES_MAP = [
            ("Access_Method", {"Access_Method", "Access Method"}),
            ("Data_Type", {"Data Type", "Data_Type"}),
            ("Platform", {"Platform"}),
        ]

        @property
        def dimensions_sources(self) -> typing.Dict[str, DimensionSource]:
            sources = super().dimensions_sources
            # Set platform as first col after header
            sources["Platform"] = DimensionSource(
                "Platform",
                CoordRange(Coord(self.header_row[0].row + 1, 0), Direction.DOWN),
            )
            return sources

    areas = [Area]


class TR(BaseCounter5Parser):
    data_format = DataFormatDefinition(name="TR")

    titles_to_skip: typing.List[str] = ["Total", "All Titles", ""]

    platforms = ["*"]
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & (
            RegexCondition(re.compile(r"^TR$"), Coord(1, 1))
            | RegexCondition(re.compile(r"^TR_B1$"), Coord(1, 1))
        )
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5(\.0)?$"), Coord(2, 1))
    )

    class Area(Counter5HeaderArea):
        DIMENSION_NAMES_MAP = [
            ("Access_Type", {"Access Type", "Access_Type"}),
            ("Access_Method", {"Access_Method", "Access Method"}),
            ("Data_Type", {"Data Type", "Data_Type"}),
            ("Section_Type", {"Section Type", "Section_Type"}),
            ("YOP", {"YOP", "Year of Publication", "Year_of_Publication"}),
            ("Publisher", {"Publisher"}),
            ("Platform", {"Platform"}),
        ]

    areas = [Area]


class IR_M1(BaseCounter5Parser):
    data_format = DataFormatDefinition(name="IR_M1")

    titles_to_skip: typing.List[str] = ["Total", "All", ""]

    platforms = ["*"]
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^IR_M1$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5(\.0)?$"), Coord(2, 1))
    )

    class Area(Counter5HeaderArea):
        title_source = None
        title_ids_sources = {}
        DIMENSION_NAMES_MAP = [
            ("Publisher", {"Publisher"}),
            ("Platform", {"Platform"}),
        ]
        ITEM_COLUMN_NAMES = ["Title", "Item"]

        ITEM_DOI_NAMES = Counter5HeaderArea.TITLE_DOI_NAMES
        ITEM_ISBN_NAMES = Counter5HeaderArea.TITLE_ISBN_NAMES
        ITEM_ISSN_NAMES = Counter5HeaderArea.TITLE_ISSN_NAMES
        ITEM_EISSN_NAMES = Counter5HeaderArea.TITLE_EISSN_NAMES
        ITEM_URI_NAMES = Counter5HeaderArea.TITLE_URI_NAMES
        ITEM_PROPRIETARY_NAMES = Counter5HeaderArea.TITLE_PROPRIETARY_NAMES

    areas = [Area]


class IR(BaseCounter5Parser):
    data_format = DataFormatDefinition(name="IR")

    titles_to_skip: typing.List[str] = []
    items_to_skip: typing.List[str] = ["Total", "All"]

    platforms = ["*"]
    heuristics = (
        RegexCondition(re.compile(r"^Report_Name$"), Coord(0, 0))
        & RegexCondition(re.compile(r"^Report_ID$"), Coord(1, 0))
        & RegexCondition(re.compile(r"^IR$"), Coord(1, 1))
        & RegexCondition(re.compile(r"^Release$"), Coord(2, 0))
        & RegexCondition(re.compile(r"^5(\.0)?$"), Coord(2, 1))
    )

    class Area(Counter5HeaderArea):
        DIMENSION_NAMES_MAP = [
            ("Article_Version", {"Article Version", "Article_Version"}),
            ("Access_Type", {"Access Type", "Access_Type"}),
            ("Access_Method", {"Access_Method", "Access Method"}),
            ("Data_Type", {"Data Type", "Data_Type"}),
            ("Parent_Data_Type", {"Parent Data Type", "Parent_Data_Type"}),
            ("YOP", {"YOP", "Year of Publication", "Year_of_Publication"}),
            ("Platform", {"Platform"}),
            ("Publisher", {"Publisher"}),
        ]
        TITLE_COLUMN_NAMES = ["Parent_Title"]
        ITEM_COLUMN_NAMES = ["Title", "Item"]

        ITEM_DOI_NAMES = Counter5HeaderArea.TITLE_DOI_NAMES
        ITEM_ISBN_NAMES = Counter5HeaderArea.TITLE_ISBN_NAMES
        ITEM_ISSN_NAMES = Counter5HeaderArea.TITLE_ISSN_NAMES
        ITEM_EISSN_NAMES = Counter5HeaderArea.TITLE_EISSN_NAMES
        ITEM_URI_NAMES = Counter5HeaderArea.TITLE_URI_NAMES
        ITEM_PROPRIETARY_NAMES = Counter5HeaderArea.TITLE_PROPRIETARY_NAMES
        ITEM_AUTHORS_NAMES = {"Authors"}
        ITEM_PUBLICATION_DATE_NAMES = {"Publication_Date"}

        TITLE_DOI_NAMES = {"Parent_DOI"}
        TITLE_ISBN_NAMES = {"Parent_ISBN"}
        TITLE_ISSN_NAMES = {"Parent_Print_ISSN"}
        TITLE_EISSN_NAMES = {"Parent_Online_ISSN"}
        TITLE_URI_NAMES = {"Parent_URI"}
        TITLE_PROPRIETARY_NAMES = {"Parent_Proprietary_ID"}

    areas = [Area]



import itertools
import logging
import typing
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from dataclasses import asdict, dataclass, field, fields
from datetime import date
from enum import Enum

from celus_nigiri import CounterRecord
from pydantic import Field
from pydantic.dataclasses import dataclass as pydantic_dataclass
from pydantic.dataclasses import rebuild_dataclass
from typing_extensions import Annotated

from celus_nibbler.conditions import Condition
from celus_nibbler.coordinates import Coord, CoordRange, Direction
from celus_nibbler.errors import TableException
from celus_nibbler.reader import SheetReader
from celus_nibbler.sources import (
    DateSource,
    DimensionSource,
    ExtractParams,
    MetricSource,
    OrganizationSource,
    TitleIdSource,
    TitleSource,
    ValueSource,
    VoidSource,
)
from celus_nibbler.utils import JsonEncorder, PydanticConfig, end_month, start_month

Role = Annotated[
    typing.Union[
        DimensionSource,
        MetricSource,
        OrganizationSource,
        TitleSource,
        TitleIdSource,
        DateSource,
        VoidSource,
        ValueSource,
    ],
    Field(discriminator="role"),
]


COUNTER_RECORD_FIELD_NAMES = [f.name for f in fields(CounterRecord)]
COUNTER_RECORD_FIELD_NAMES_DATES = ["start", "end"]
COUNTER_RECORD_FIELD_NAMES_SIMPLE = [
    f
    for f in COUNTER_RECORD_FIELD_NAMES
    if f not in ("value", "title_ids", "dimension_data", "start", "end")
]
COUNTER_RECORD_FIELD_NAMES_NESTED = ["title_ids", "dimension_data"]


logger = logging.getLogger(__name__)


# Make sure that wrong definition doesn't create
# an infinite loop while processing header
MAX_DATA_CELLS = 1000

# Make sure that wrong definition doesn't create
# an infinite loop while looking for header
MAX_HEADER_OFFSET_LOOKUP_COUNT = 1_000_000


class DataHeaderBaseCondition(metaclass=ABCMeta):
    @abstractmethod
    def check(self, value: str, idx: int):
        pass


class ArithmeticsMixin:
    def __invert__(self):
        return DataHeaderNegCondition(self)

    def __or__(self, other):
        return DataHeaderOrCondition([self, other])

    def __and__(self, other):
        return DataHeaderAndCondition([self, other])


@pydantic_dataclass(config=PydanticConfig)
class DataHeaderNegCondition(ArithmeticsMixin, DataHeaderBaseCondition, JsonEncorder):
    cond: "DataHeaderCondition"

    kind: typing.Literal["neg"] = "neg"

    def check(self, *args, **kwargs):
        return not self.cond.check(*args, **kwargs)


@pydantic_dataclass(config=PydanticConfig)
class DataHeaderAndCondition(ArithmeticsMixin, DataHeaderBaseCondition, JsonEncorder):
    conds: typing.List["DataHeaderCondition"]

    kind: typing.Literal["and"] = "and"

    def check(self, *args, **kwargs):
        return all(e.check(*args, **kwargs) for e in self.conds)


@pydantic_dataclass(config=PydanticConfig)
class DataHeaderOrCondition(ArithmeticsMixin, DataHeaderBaseCondition, JsonEncorder):
    conds: typing.List["DataHeaderCondition"]

    kind: typing.Literal["or"] = "or"

    def check(self, *args, **kwargs):
        return any(e.check(*args, **kwargs) for e in self.conds)


@pydantic_dataclass(config=PydanticConfig)
class DataHeaderRegexCondition(JsonEncorder):
    pattern: typing.Pattern

    kind: typing.Literal["regex"] = "regex"

    def check(self, value: typing.Any, idx: int) -> bool:
        if isinstance(value, str):
            return bool(self.pattern.match(value))

        return False


@pydantic_dataclass(config=PydanticConfig)
class DataHeaderIndexCondition(JsonEncorder):
    max: typing.Optional[int]
    min: typing.Optional[int]

    kind: typing.Literal["index"] = "index"

    def check(self, value: typing.Any, idx: int) -> bool:
        max = self.max if self.max is not None else idx
        min = self.min if self.min is not None else 0
        return min <= idx <= max


DataHeaderCondition = Annotated[
    typing.Union[
        DataHeaderAndCondition,
        DataHeaderNegCondition,
        DataHeaderOrCondition,
        DataHeaderRegexCondition,
        DataHeaderIndexCondition,
    ],
    Field(discriminator="kind"),
]


rebuild_dataclass(DataHeaderNegCondition, force=True)
rebuild_dataclass(DataHeaderAndCondition, force=True)
rebuild_dataclass(DataHeaderOrCondition, force=True)


class DataHeaderAction(Enum):
    PROCEED = "proceed"
    BYPASS = "bypass"
    SKIP = "skip"
    STOP = "stop"

    def merge(self, dha: "DataHeaderAction") -> "DataHeaderAction":
        if self == DataHeaderAction.PROCEED:
            return dha
        elif self == DataHeaderAction.BYPASS:
            return dha if dha != DataHeaderAction.PROCEED else self
        elif self == DataHeaderAction.SKIP:
            if dha != DataHeaderAction.PROCEED:
                return dha

        return self


@pydantic_dataclass(config=PydanticConfig)
class DataHeaderRule(JsonEncorder):
    condition: typing.Optional[DataHeaderCondition] = None
    on_condition_failed: DataHeaderAction = DataHeaderAction.STOP
    on_condition_passed: DataHeaderAction = DataHeaderAction.PROCEED
    on_error: DataHeaderAction = DataHeaderAction.STOP
    role_idx: typing.Optional[int] = None
    role_extract_params_override: typing.Optional[ExtractParams] = None
    role_source_offset: int = 0

    def process(
        self,
        sheet: SheetReader,
        idx: int,
        role: Role,
        parser_row_offset: typing.Optional[int],
        area_row_offset: typing.Optional[int],
    ) -> typing.Tuple[DataHeaderAction, typing.Optional[typing.Any]]:
        if role.cleanup_during_header_processing:
            role = deepcopy(role)

        if self.role_extract_params_override:
            role.extract_params = deepcopy(self.role_extract_params_override)

        try:
            value = role.extract(
                sheet,
                idx + self.role_source_offset,
                parser_row_offset=parser_row_offset,
                area_row_offset=area_row_offset,
            )
            if self.condition is None or self.condition.check(value, idx):
                return self.on_condition_passed, value
            else:
                return self.on_condition_failed, None

        except TableException as e:
            if e.action == TableException.Action.STOP:
                # terminate if no offset
                action = (
                    DataHeaderAction.STOP
                    if self.role_source_offset == 0
                    else DataHeaderAction.PROCEED
                )
            else:
                action = self.on_error

            logger.debug(
                "Header role %s processing rule %s failed (->%s): %s", role, self, action, e
            )
            return action, None


@pydantic_dataclass(config=PydanticConfig)
class DataHeaders(JsonEncorder):
    roles: typing.List[Role]

    data_cells: CoordRange  # first data after the header
    data_direction: Direction  # perpendicular to data_cells
    data_extract_params: ExtractParams = field(
        default_factory=lambda: ExtractParams(on_validation_error=TableException.Action.STOP)
    )
    data_allow_negative: bool = False
    data_cells_options: "DataCellsOptions" = field(
        default_factory=lambda: DataCellsOptions(use_header_year=True, use_header_month=True)
    )

    rules: typing.List[DataHeaderRule] = Field(default_factory=lambda: [DataHeaderRule()])
    condition: typing.Optional[Condition] = None

    def process_value(
        self,
        record: CounterRecord,
        role: Role,
        value: typing.Any,
        get_metric_name: typing.Callable[[str], str],
        check_metric_name: typing.Callable[[str, str], None],
    ) -> DataHeaderAction:
        if isinstance(role, DimensionSource):
            record.dimension_data[role.name] = value
        elif isinstance(role, TitleIdSource):
            if value.strip():
                record.title_ids[role.name] = value
        elif isinstance(role, DateSource):
            record.start = start_month(value)
            record.end = end_month(value)
        elif isinstance(role, MetricSource):
            metric_name = get_metric_name(value)
            try:
                check_metric_name(metric_name, value)
            except TableException as e:
                if e.action == TableException.Action.SKIP:
                    return DataHeaderAction.SKIP
                elif e.action == TableException.Action.STOP:
                    return DataHeaderAction.STOP
                raise
            record.metric = metric_name
        elif isinstance(role, VoidSource):
            pass
        else:
            setattr(record, role.role, value)
        return DataHeaderAction.PROCEED

    def prepare_row_offset(
        self, sheet: SheetReader, parser_row_offset: typing.Optional[int]
    ) -> int:
        parser_row_offset = parser_row_offset or 0
        if not self.condition:
            # No condition specified -> use initial
            return parser_row_offset

        # Iterate until condition matches or an exception is raised
        for area_offset in range(
            parser_row_offset, min(MAX_HEADER_OFFSET_LOOKUP_COUNT + parser_row_offset, len(sheet))
        ):
            try:
                if self.condition.check(sheet, parser_row_offset, area_offset):
                    return area_offset

            except TableException as e:
                raise TableException(
                    row=None, col=None, sheet=e.sheet, reason="no-header-data-found"
                ) from e

        # no offset found
        raise TableException(
            row=None, col=None, sheet=sheet.sheet_idx, reason="no-header-data-found"
        )

    def detect_data_cells(
        self,
        sheet: SheetReader,
        parser_row_offset: typing.Optional[int],
        get_metric_name: typing.Callable[[str], str],
        check_metric_name: typing.Callable[[str, str], None],
    ) -> typing.Tuple[int, typing.List["DataCells"]]:
        res: typing.List[DataCells] = []

        # Derive row offset
        area_row_offset = self.prepare_row_offset(sheet, parser_row_offset)

        # When role with ValueSource is present make it value_source
        # for all other roles
        root_value_source = None
        for role in self.roles:
            if isinstance(role, ValueSource):
                res.append(
                    DataCells(
                        header_data=CounterRecord(value=0),
                        value_source=role,
                        options=deepcopy(self.data_cells_options),
                    )
                )
                break

        for idx, cell in itertools.takewhile(
            lambda x: x[0] < MAX_DATA_CELLS, enumerate(self.data_cells)
        ):
            record = CounterRecord(value=0)
            action = DataHeaderAction.PROCEED
            store = False
            value = None

            value_source = ValueSource(
                source=CoordRange(cell, self.data_direction),
                extract_params=self.data_extract_params,
                allow_negative=self.data_allow_negative,
            )

            # Process data from header
            for role_idx, role in enumerate(self.roles):
                if isinstance(role, ValueSource):
                    continue

                for rule_idx, rule in enumerate(self.rules):
                    if rule.role_idx and role_idx != rule.role_idx:
                        # Skip rules which doesn't match index
                        continue
                    cur_action, value = rule.process(
                        sheet, idx, role, parser_row_offset, area_row_offset
                    )
                    action = action.merge(cur_action)

                    if action != DataHeaderAction.PROCEED or value is not None:
                        break

                if action not in [DataHeaderAction.PROCEED, DataHeaderAction.BYPASS]:
                    # Terminate processing of other roles
                    break

                # merge with action which occured during processing
                if value and action == DataHeaderAction.PROCEED:
                    action = action.merge(
                        self.process_value(record, role, value, get_metric_name, check_metric_name)
                    )

                if action not in [DataHeaderAction.PROCEED, DataHeaderAction.BYPASS]:
                    # Terminate processing of other roles
                    break

                if value and action == DataHeaderAction.PROCEED:
                    store = True

            if action == DataHeaderAction.SKIP:
                logger.debug("Header parsing skips: %s", cell)
                continue

            if action == DataHeaderAction.STOP:
                logger.debug("Header parsing stops: %s", cell)
                break

            if store:
                res.append(
                    DataCells(
                        record,
                        root_value_source or value_source,
                        options=deepcopy(self.data_cells_options),
                    )
                )

        if not res and len(self.roles) > 0:
            raise TableException(
                row=self.data_cells.coord.row_absolute(parser_row_offset, area_row_offset),
                col=self.data_cells.coord.col,
                sheet=sheet.sheet_idx,
                reason="no-header-data-found",
            )

        return area_row_offset, res


@dataclass
class DataCellsOptions:
    use_header_year: bool = True
    use_header_month: bool = True


@dataclass
class DataCells:
    header_data: CounterRecord
    value_source: ValueSource
    options: DataCellsOptions = field(default_factory=lambda: DataCellsOptions(False, False))

    def __iter__(self):
        return self.value_source.source.__iter__()

    def __next__(self):
        return self.value_source.source.__next__()

    def __getitem__(self, item: int) -> Coord:
        if isinstance(self.value_source.source, CoordRange):
            return self.value_source.source.__getitem__(item)
        else:
            raise TypeError()

    def __str__(self):
        records = {f"({k}={v})" for k, v in asdict(self.header_data).items() if v}
        return f"{'|'.join(records)} - {self.value_source}"

    def __repr__(self):
        return str(self)

    def merge_into_record(
        self,
        record: CounterRecord,
    ) -> CounterRecord:
        # Update the input record instead creating new one
        # to avoid unnecessary alocation

        # deal with updating date fields
        # not that it can be updated partially (year from row, month from header)
        if self.options.use_header_year and self.options.use_header_month:
            for field_name in COUNTER_RECORD_FIELD_NAMES_DATES:
                if new_value := getattr(self.header_data, field_name):
                    setattr(record, field_name, new_value)
        else:
            for field_name in COUNTER_RECORD_FIELD_NAMES_DATES:
                if record_date := getattr(record, field_name):
                    if header_date := getattr(self.header_data, field_name):
                        h_year = header_date.year
                        h_month = header_date.month
                        r_year = record_date.year
                        r_month = record_date.month
                        new_date = date(
                            h_year if self.options.use_header_year else r_year,
                            h_month if self.options.use_header_month else r_month,
                            1,
                        )
                        if field_name == "end":
                            new_date = end_month(new_date)
                        setattr(record, field_name, new_date)
                else:
                    # record date empty use date from header regardless of use_header_*
                    if new_value := getattr(self.header_data, field_name):
                        setattr(record, field_name, new_value)

        # Update non-nested updatable fields
        for field_name in COUNTER_RECORD_FIELD_NAMES_SIMPLE:
            if new_value := getattr(self.header_data, field_name):
                setattr(record, field_name, new_value)

        # Update nested fields
        for nested_name in COUNTER_RECORD_FIELD_NAMES_NESTED:
            if header_nested_data := getattr(self.header_data, nested_name):
                if record_nested_data := getattr(record, nested_name):
                    record_nested_data.update(header_nested_data)
                else:
                    # Clone header data
                    setattr(record, nested_name, deepcopy(header_nested_data))

        return record


@pydantic_dataclass(config=PydanticConfig)
class DataFormatDefinition(JsonEncorder):
    name: str
    id: typing.Optional[int] = None



import typing

from pydantic import Field, RootModel
from typing_extensions import Annotated

from ..utils import JsonEncorder
from . import counter
from .base import BaseParserDefinition
from .celus_format import CelusFormatAreaDefinition, CelusFormatParserDefinition
from .counter import BaseCounterAreaDefinition, CounterDefinition
from .generic import GenericAreaDefinition, GenericDefinition

AreaDefinition = Annotated[
    typing.Union[
        counter.BR1AreaDefinition,
        counter.BR2AreaDefinition,
        counter.BR3AreaDefinition,
        counter.DB1AreaDefinition,
        counter.DB2AreaDefinition,
        counter.JR1AreaDefinition,
        counter.JR1aAreaDefinition,
        counter.JR1GOAAreaDefinition,
        counter.JR2AreaDefinition,
        counter.MR1AreaDefinition,
        counter.PR1AreaDefinition,
        counter.TRAreaDefinition,
        counter.IR_M1AreaDefinition,
        counter.IRAreaDefinition,
        counter.DRAreaDefinition,
        counter.PRAreaDefinition,
        counter.TR51AreaDefinition,
        counter.IR51AreaDefinition,
        counter.DR51AreaDefinition,
        counter.PR51AreaDefinition,
        CelusFormatAreaDefinition,
        GenericAreaDefinition,
    ],
    Field(discriminator="kind"),
]


DefinitionAnotation = Annotated[
    typing.Union[
        counter.BR1Definition,
        counter.BR2Definition,
        counter.BR3Definition,
        counter.DB1Definition,
        counter.DB2Definition,
        counter.JR1Definition,
        counter.JR1aDefinition,
        counter.JR1GOADefinition,
        counter.JR2Definition,
        counter.MR1Definition,
        counter.PR1Definition,
        counter.DRDefinition,
        counter.PRDefinition,
        counter.TRDefinition,
        counter.IR_M1Definition,
        counter.IRDefinition,
        counter.TR51Definition,
        counter.IR51Definition,
        counter.DR51Definition,
        counter.PR51Definition,
        CelusFormatParserDefinition,
        GenericDefinition,
    ],
    Field(discriminator="kind"),
]


def get_all_definitions() -> typing.List[BaseParserDefinition]:
    return list(DefinitionAnotation.__origin__.__args__)


class Definition(JsonEncorder, RootModel):
    root: DefinitionAnotation


__all__ = [
    "AreaDefinition",
    "BaseCounterAreaDefinition",
    "CounterDefinition",
    "Definition",
    "BaseParserDefinition",
    "GenericDefinition",
    "GenericAreaDefinition",
    "CelusFormatAreaDefinition",
    "CelusFormatParserDefinition",
]
